import anndata
import pandas as pd


class CompositeGLMSimulator:
    def __init__(self, specification: dict, **kwargs):
        self.specification = specification
        self.params = {}
        self.hyperparams = kwargs

        for k in self.specification:
            if "_fitted" not in self.specification[k].keys():
                self.specification[k]["_fitted"] = False


    def fit(self, adata: anndata.AnnData) -> dict:
        self.specification = fill_var_names(self.specification, list(adata.var_names))

        for k, spec in self.specification.items():
            if not spec["_fitted"]:
                spec["simulator"].fit(adata[:, spec["var_names"]], spec["formula"])
            self.params[k] = subset_params(spec["simulator"].params, list(spec["var_names"]))
            self.specification[k]["simulator"].params = self.params[k]
            self.specification[k]["_fitted"] = True

    def sample(self, obs: pd.DataFrame) -> anndata.AnnData:
        anndata_list = []
        for spec in self.specification.values():
            anndata_list.append(spec["simulator"].sample(obs))
        return anndata.concat(anndata_list, axis="var")

    def predict(self, obs: pd.DataFrame) -> dict:
        preds = {}
        for k, spec in self.specification.items():
            preds[k] = spec["simulator"].predict(obs)
        return preds

    def __repr__(self):
        var_names = {k: list_string(v["var_names"]) for k, v in self.specification.items()}
        simulators = {k: v["simulator"] for k, v in self.specification.items()}
        return f"""scDesigner simulator object with
    method: 'Composite'
    features: {var_names}
    simulators: {simulators}"""


def list_string(l):
    if l is None:
        return
    if len(l) <= 3:
        return ", ".join(l)
    return f"[{l[0]},{l[1]}, ..., {l[-1]}]"

def fill_var_names(specification, var_names):
    all_names = []
    for k, v in specification.items():
        if v["var_names"] is not None:
            all_names += list(v["var_names"])

    for k, v in specification.items():
        if v["var_names"] is None:
            specification[k]["var_names"] = list(set(var_names) - set(all_names))
    return specification

def subset_params(params, var_names):
    result = {}
    for k, v in params.items():
        if "covariance" in k:
            result[k] = v.loc[var_names, var_names]
        else:
            result[k] = v.loc[:, var_names]
    return result


from anndata import AnnData
from formulaic import model_matrix
from torch.utils.data import DataLoader, TensorDataset
from typing import Union
import numpy as np
import pandas as pd
import scipy.sparse
import torch
import torch.utils.data as td
import warnings

def formula_loader(
    adata: AnnData, formula=None, chunk_size=int(1e4), batch_size: int = None
):
    device = check_device()
    if adata.isbacked:
        ds = FormulaViewDataset(adata, formula, chunk_size, device)
        dataloader = td.DataLoader(ds, batch_size=batch_size)
        ds.x_names = model_matrix_names(adata, formula, ds.categories)
    else:
        # convert sparse to dense matrix
        y = adata.X
        if isinstance(y, scipy.sparse._csc.csc_matrix):
            y = y.todense()

        # create tensor-based loader
        x = model_matrix(formula, pd.DataFrame(adata.obs))
        ds = TensorDataset(
            torch.tensor(np.array(x), dtype=torch.float32).to(device),
            torch.tensor(y, dtype=torch.float32).to(device),
        )
        ds.x_names = list(x.columns)
        dataloader = DataLoader(ds, batch_size=batch_size, shuffle=False)

    return dataloader

def multiple_formula_loader(
    adata: AnnData, formulas: dict, chunk_size=int(1e4), batch_size: int = None
):
    dataloaders = {}
    for key in formulas.keys():
        dataloaders[key] = formula_loader(adata, formulas[key], chunk_size, batch_size)
    return dataloaders

class FormulaViewDataset(td.Dataset):
    def __init__(self, view, formula=None, chunk_size=int(1e4), device=None):
        super().__init__()
        self.view = view
        self.formula = formula
        self.len = len(view)
        self.cur_range = range(0, min(self.len, chunk_size))
        self.categories = column_levels(view.obs)
        self.x = None
        self.y = None
        self.device = device or check_device()

    def __len__(self):
        return self.len

    def __getitem__(self, ix):
        if self.x is None or ix not in self.cur_range:
            self.cur_range = range(ix, min(ix + len(self.cur_range), self.len))
            view_inmem = self.view[self.cur_range].to_memory()
            self.x = safe_model_matrix(
                view_inmem.obs, self.formula, self.categories
            ).to(self.device)
            self.y = torch.from_numpy(view_inmem.X.toarray().astype(np.float32)).to(
                self.device
            )
        return self.x[ix - self.cur_range[0]], self.y[ix - self.cur_range[0]]


def replace_cols(obs, categories):
    for k in obs.columns:
        if str(obs[k].dtype) == "category":
            obs[k] = obs[k].astype(pd.CategoricalDtype(categories[k]))
    return obs


def model_matrix_names(adata, formula, categories):
    if adata.isbacked:
        obs = adata[:1].to_memory().obs
    else:
        obs = adata.obs

    if formula is None:
        return list(obs.columns)

    obs = replace_cols(obs, categories)
    return list(model_matrix(formula, pd.DataFrame(obs)).columns)


def safe_model_matrix(obs, formula, categories):
    if formula is None:
        return obs

    obs = replace_cols(obs, categories)
    x = model_matrix(formula, pd.DataFrame(obs))
    return torch.from_numpy(np.array(x).astype(np.float32))


def column_levels(obs):
    categories = {}
    for k in obs.columns:
        obs_type = str(obs[k].dtype)
        if obs_type in ["category", "object"]:
            categories[k] = obs[k].unique()
    return categories


def check_device():
    return torch.device(
        "cuda"
        if torch.cuda.is_available()
        else "mps" if torch.backends.mps.is_available() else "cpu"
    )

def standardize_formula(formula: Union[str, dict], allowed_keys = None):
    # The first element of allowed_keys should be the name of default parameter
    if allowed_keys is None:
        raise ValueError("Internal error: allowed_keys must be specified")
    formula = {allowed_keys[0]: formula} if isinstance(formula, str) else formula

    formula_keys = set(formula.keys())
    allowed_keys = set(allowed_keys)

    if not formula_keys & allowed_keys:
        raise ValueError(f"formula must have at least one of the following keys: {allowed_keys}")
    
    if extra_keys := formula_keys - allowed_keys:
        warnings.warn(
            f"Invalid formulas in dictionary will not be used: {extra_keys}",
            UserWarning,
        )
    
    formula.update({k: '~ 1' for k in allowed_keys - formula_keys})
    return formula


from .. import estimators as est
from .. import predictors as prd
from .. import samplers as smp
from .. import diagnose
from anndata import AnnData
import pandas as pd
from typing import Union


def glm_simulator_generator(class_name, regressor, sampler, predictor, diagnose):
    def __init__(self, **kwargs):
        self.formula = None # formula should be a string or a dictionary of strings
        self.params = None
        self.marginal_aic = None
        self.marginal_bic = None
        self.copula_aic = None
        self.copula_bic = None
        self.hyperparams = kwargs
        self.filtered_kwargs = {k: kwargs[k] for k in ["chunk_size", "batch_size"] if k in kwargs}

    if "Copula" not in class_name:
        # fitting and sampling methods for plain regressors
        def fit(self, adata: AnnData, formula: Union[str, dict]) -> dict:
            self.formula = formula
            self.params = regressor(adata, formula, **self.hyperparams)
            self.marginal_aic, self.marginal_bic = diagnose(
                self.params, adata, formula, **self.filtered_kwargs
            )

        def sample(self, obs: pd.DataFrame) -> AnnData:
            local_parameters = self.predict(obs) # a dictionary of parameters
            return sampler(local_parameters, obs)

    else:
        # fitting and sampling for gaussian copula models

        def fit(
            self, adata: AnnData, formula: Union[str, dict] = "~ 1", copula_groups: str = None
        ) -> dict:
            self.formula = formula
            self.copula_groups = copula_groups
            self.params = regressor(adata, formula, copula_groups, **self.hyperparams)
            self.marginal_aic, self.marginal_bic, self.copula_aic, self.copula_bic = diagnose(
                self.params, adata, formula, copula_groups, **self.filtered_kwargs
            )

        def sample(self, obs: pd.DataFrame) -> AnnData:
            groups = est.gaussian_copula_factory.group_indices(self.copula_groups, obs)
            local_parameters = self.predict(obs)
            return sampler(local_parameters, self.params["covariance"], groups, obs)

    def predict(self, obs: pd.DataFrame) -> dict:
        return predictor(self.params, obs, self.formula)
        # The predictor function should handle different formula types: dict or string

    def __repr__(self):
        params_string = ", ".join(
            [
                f"{k} [{self.params[k].shape[0]}x{self.params[k].shape[1]}]"
                for k in self.params.keys()
            ]
        )
        return f"""scDesigner simulator object with
    method: {self.__class__.__name__}
    formula: {self.formula}
    parameters: {params_string}"""

    return type(
        class_name,
        (),
        {
            "__init__": __init__,
            "fit": fit,
            "sample": sample,
            "predict": predict,
            "__repr__": __repr__,
        },
    )


NegBinRegressionSimulator = glm_simulator_generator(
    "NegBinRegressionSimulator",
    est.negbin_regression,
    smp.negbin_sample,
    prd.negbin_predict,
    diagnose.negbin_regression_diagnose
)

NegBinCopulaSimulator = glm_simulator_generator(
    "NegBinCopulaSimulator",
    est.negbin_copula, 
    smp.negbin_copula_sample,
    prd.negbin_predict,
    diagnose.negbin_gcopula_diagnose,
)

PoissonRegressionSimulator = glm_simulator_generator(
    "PoissonRegressionSimulator",
    est.poisson_regression,
    smp.poisson_sample,
    prd.poisson_predict,
    diagnose.poisson_regression_diagnose,
)

PoissonCopulaSimulator = glm_simulator_generator(
    "PoissonCopulaSimulator",
    est.poisson_copula,
    smp.poisson_copula_sample,
    prd.poisson_predict,
    diagnose.poisson_gcopula_diagnose,
)

BernoulliRegressionSimulator = glm_simulator_generator(
    "BernoulliRegressionSimulator",
    est.bernoulli_regression,
    smp.bernoulli_sample,
    prd.bernoulli_predict,
    diagnose.bernoulli_regression_diagnose,
)
 
BernoulliCopulaSimulator = glm_simulator_generator(
    "BernoulliCopulaSimulator",
    est.bernoulli_copula,
    smp.bernoulli_copula_sample,
    prd.bernoulli_predict,
    diagnose.bernoulli_gcopula_diagnose,
)

ZeroInflatedNegBinRegressionSimulator = glm_simulator_generator(
    "ZeroInflatedNegbinRegressionSimulator",
    est.zero_inflated_negbin_regression,
    smp.zero_inflated_negbin_sample,
    prd.zero_inflated_negbin_predict,
    diagnose.zinb_regression_diagnose,
)

ZeroInflatedNegBinCopulaSimulator = glm_simulator_generator(
    "ZeroInflatedNegBinCopulaSimulator",
    est.zero_inflated_negbin_copula,
    smp.zero_inflated_negbin_copula_sample,
    prd.zero_inflated_negbin_predict,
    diagnose.zinb_gcopula_diagnose,
)

ZeroInflatedPoissonRegressionSimulator = glm_simulator_generator(
    "ZeroInflatedNegbinRegressionSimulator",
    est.zero_inflated_poisson_regression,
    smp.zero_inflated_poisson_sample,
    prd.zero_inflated_poisson_predict,
    diagnose.zip_regression_diagnose,
)

GaussianRegressionSimulator = glm_simulator_generator(
    "GaussianRegressionSimulator",
    est.gaussian_regression,
    smp.gaussian_regression_sample,
    prd.gaussian_predict,
    diagnose.gaussian_regression_diagnose
)

GaussianCopulaSimulator = glm_simulator_generator(
    "GaussianCopulaSimulator",
    est.gaussian_copula,
    smp.gaussian_copula_sample,
    prd.gaussian_predict,
    diagnose.gaussian_gcopula_diagnose
)



from scipy.sparse.linalg import svds
from sklearn.cluster import KMeans
import numpy as np
import pandas as pd
import torch


# computes PNMF weight and score, ncol specify the number of clusters
def pnmf(log_data, nbase=3, **kwargs):  # data is np array, log transformed read data
    """
    Computes PNMF weight and score.

    :log_data: log transformed np array of read data
    :ncol: specify the number of clusters
    :return: W (weights, gene x base) and S (scores, base x cell) as numpy arrays
    """
    U = left_singular(log_data, nbase)
    W = pnmf_eucdist(log_data, U, **kwargs)
    W = W / np.linalg.norm(W, ord=2)
    S = W.T @ log_data
    return W, S


def gamma_regression_array(
    x: np.array, y: np.array, batch_size: int = 512, lr: float = 0.1, epochs: int = 40
) -> dict:
    x = torch.tensor(x, dtype=torch.float32)
    y = torch.tensor(y, dtype=torch.float32)

    n_features, n_outcomes = x.shape[1], y.shape[1]
    a = torch.zeros(n_features * n_outcomes, requires_grad=True)
    loc = torch.zeros(n_features * n_outcomes, requires_grad=True)
    beta = torch.zeros(n_features * n_outcomes, requires_grad=True)
    optimizer = torch.optim.Adam([a, loc, beta], lr=lr)

    for i in range(epochs):
        optimizer.zero_grad()
        loss = negative_gamma_log_likelihood(a, beta, loc, x, y)
        loss.backward()
        optimizer.step()

    a = to_np(a).reshape(n_features, n_outcomes)
    loc = to_np(loc).reshape(n_features, n_outcomes)
    beta = to_np(beta).reshape(n_features, n_outcomes)
    return {"a": a, "loc": loc, "beta": beta}


def class_generator(score, n_clusters=3):
    """
    Generates one-hot encoding for score classes
    """
    kmeans = KMeans(n_clusters, random_state=0)  # Specify the number of clusters
    kmeans.fit(score.T)
    labels = kmeans.labels_
    num_classes = len(np.unique(labels))
    one_hot = np.eye(num_classes)[labels].astype(int)
    return labels


###############################################################################
## Helpers for deriving PNMF
###############################################################################


def pnmf_eucdist(X, W_init, maxIter=500, threshold=1e-4, tol=1e-10, verbose=False):
    # initialization
    W = W_init  # initial W is the PCA of X
    XX = X @ X.T

    # iterations
    for iter in range(maxIter):
        if verbose and (iter + 1) % 10 == 0:
            print("%d iterations used." % (iter + 1))
        W_old = W

        XXW = XX @ W
        SclFactor = np.dot(W, W.T @ XXW) + np.dot(XXW, W.T @ W)

        # QuotientLB
        SclFactor = MatFindlb(SclFactor, tol)
        SclFactor = XXW / SclFactor
        W = W * SclFactor  # somehow W *= SclFactor doesn't work?

        norm_W = np.linalg.norm(W)
        W /= norm_W
        W = MatFind(W, tol)

        diffW = np.linalg.norm(W_old - W) / np.linalg.norm(W_old)
        if diffW < threshold:
            break

    return W


# left singular vector of X
def left_singular(X, k):
    U, _, _ = svds(X, k=k)
    return np.abs(U)


def MatFindlb(A, lb):
    B = np.ones(A.shape) * lb
    Alb = np.where(A < lb, B, A)
    return Alb


def MatFind(A, ZeroThres):
    B = np.zeros(A.shape)
    Atrunc = np.where(A < ZeroThres, B, A)
    return Atrunc


###############################################################################
## Helpers for training PNMF regression
###############################################################################


def shifted_gamma_pdf(x, alpha, beta, loc):
    if not torch.is_tensor(x):
        x = torch.tensor(x)
    mask = x < loc
    y_clamped = torch.clamp(x - loc, min=1e-12)

    log_pdf = (
        alpha * torch.log(beta)
        - torch.lgamma(alpha)
        + (alpha - 1) * torch.log(y_clamped)
        - beta * y_clamped
    )
    loss = -torch.mean(log_pdf[~mask])
    n_invalid = mask.sum()
    if n_invalid > 0:  # force samples to be greater than loc
        loss = loss + 1e10 * n_invalid.float()
    return loss


def negative_gamma_log_likelihood(log_a, log_beta, loc, X, y):
    n_features = X.shape[1]
    n_outcomes = y.shape[1]

    a = torch.exp(log_a.reshape(n_features, n_outcomes))
    beta = torch.exp(log_beta.reshape(n_features, n_outcomes))
    loc = loc.reshape(n_features, n_outcomes)
    return shifted_gamma_pdf(y, X @ a, X @ beta, X @ loc)


def to_np(x):
    return x.detach().cpu().numpy()


def format_gamma_parameters(
    parameters: dict,
    W_index: list,
    coef_index: list,
) -> dict:
    parameters["a"] = pd.DataFrame(parameters["a"], index=coef_index)
    parameters["loc"] = pd.DataFrame(parameters["loc"], index=coef_index)
    parameters["beta"] = pd.DataFrame(parameters["beta"], index=coef_index)
    parameters["W"] = pd.DataFrame(parameters["W"], index=W_index)
    return parameters



from . import glm_factory as glm
from scipy.stats import poisson
from typing import Union
import numpy as np


def poisson_regression_sample_array(local_parameters: dict) -> np.array:
    mu = local_parameters["mean"]
    return poisson(mu).rvs()


def poisson_copula_sample_array(
    local_parameters: dict, covariance: Union[dict, np.array], groups: dict
) -> np.array:
    # initialize uniformized gaussian samples
    N, G = local_parameters["mean"].shape
    u = glm.gaussian_copula_pseudo_obs(N, G, covariance, groups)

    # invert using poisson margins
    mu = local_parameters["mean"]
    return poisson(mu).ppf(u)


poisson_sample = glm.glm_sample_factory(poisson_regression_sample_array)
poisson_copula_sample = glm.gaussian_copula_sample_factory(poisson_copula_sample_array)



import numpy as np
import pandas as pd
from ..format import format_matrix


def bernoulli_predict(parameters: dict, obs: pd.DataFrame, formula: str):
    x = format_matrix(obs, formula)
    theta = np.exp(x @ parameters["coef_mean"])
    return {"mean": theta}



from anndata import AnnData
from formulaic import model_matrix
from ..format import format_input_anndata, format_matrix
from scipy.stats import gamma
from ..estimators.pnmf import pnmf, gamma_regression_array, format_gamma_parameters
import numpy as np
import pandas as pd


class PNMFRegressionSimulator:
    def __init__(self, nbase=20, maxIter=100, **kwargs):  # default input: cell x gene
        self.var_names = None
        self.formula = None
        self.params = None
        self.hyperparams = {"pnmf": {"nbase": nbase, "maxIter": maxIter}, "gamma": kwargs}

    def fit(self, adata, formula: str):
        adata = format_input_anndata(adata)
        self.var_names = adata.var_names
        self.formula = formula
        log_data = np.log1p(adata.X).T
        W, S = pnmf(log_data, **self.hyperparams["pnmf"])
        adata = AnnData(X=S.T, obs=adata.obs)

        x = model_matrix(formula, adata.obs)
        parameters = gamma_regression_array(np.array(x), adata.X, **self.hyperparams["gamma"])
        parameters["W"] = W
        self.params = format_gamma_parameters(
            parameters, list(self.var_names), list(x.columns)
        )

    def sample(self, obs: pd.DataFrame) -> AnnData:
        W = self.params["W"]
        params = self.predict(obs)
        a, loc, beta = params["a"], params["loc"], params["beta"]
        sim_score = gamma(a, loc, 1 / beta).rvs()
        samples = np.exp(W @ sim_score.T).T

        # thresholding samples
        floor = np.floor(samples)
        samples = floor + np.where(samples - floor < 0.9, 0, 1) - 1
        samples = np.where(samples < 0, 0, samples)

        result = AnnData(X=samples, obs=obs)
        result.var_names = self.var_names
        return result

    def predict(self, obs: pd.DataFrame) -> dict:
        x = format_matrix(obs, self.formula)
        a, loc, beta = (
            x @ np.exp(self.params["a"]),
            x @ self.params["loc"],
            x @ np.exp(self.params["beta"]),
        )
        return {"a": a, "loc": loc, "beta": beta}

    def __repr__(self):
        return f"""scDesigner simulator object with
    method: 'PNMF Regression'
    formula: '{self.formula}'
    parameters: 'a', 'loc', 'beta', 'W'"""



from copy import deepcopy
from ..simulators import CompositeGLMSimulator


def split_glm(simulator, submodel, keys=["group1", "group2"]):
    initial_model = {
        "formula": simulator.formula,
        "simulator": simulator,
        "var_names": None,
        "_fitted": True,
    }

    # default to the formula in original model
    if not "formula" in submodel:
        submodel["formula"] = simulator.formula

    # default to the original simulator type, but remove existing parameters
    if not "simulator" in submodel:
        submodel["simulator"] = deepcopy(simulator)
        submodel["simulator"].params = None

    specification = {keys[0]: initial_model, keys[1]: submodel}
    return CompositeGLMSimulator(specification)



from .bernoulli import bernoulli_predict
from .negbin import negbin_predict
from .poisson import poisson_predict
from .gaussian import gaussian_predict
from .zero_inflated_negbin import zero_inflated_negbin_predict
from .zero_inflated_poisson import zero_inflated_poisson_predict

__all__ = [
    "bernoulli_predict",
    "gaussian_predict",
    "negbin_predict",
    "poisson_predict",
    "zero_inflated_negbin_predict",
    "zero_inflated_poisson_predict",
]



import warnings
from . import gaussian_copula_factory as gcf
from .. import format
from .. import data
from . import glm_factory as factory
from anndata import AnnData
from scipy.stats import nbinom
import numpy as np
import pandas as pd
import torch
from typing import Union
from scipy.special import expit

###############################################################################
## Regression functions that operate on numpy arrays
###############################################################################


def zero_inflated_negbin_regression_likelihood(params, X_dict, y):
    # get appropriate parameter shape
    n_mean_features = X_dict["mean"].shape[1]
    n_dispersion_features = X_dict["dispersion"].shape[1]
    n_zero_inflation_features = X_dict["zero_inflation"].shape[1]
    n_outcomes = y.shape[1]

    # define the likelihood parameters
    coef_mean = params[: n_mean_features * n_outcomes].\
        reshape(n_mean_features, n_outcomes)
    coef_dispersion = params[n_mean_features * n_outcomes :\
        n_mean_features * n_outcomes + n_dispersion_features * n_outcomes].\
        reshape(n_dispersion_features, n_outcomes)
    coef_zero_inflation = params[n_mean_features * n_outcomes + \
        n_dispersion_features * n_outcomes :].\
        reshape(n_zero_inflation_features, n_outcomes)
    
    mu = torch.exp(X_dict["mean"] @ coef_mean)
    r = torch.exp(X_dict["dispersion"] @ coef_dispersion)
    pi = torch.sigmoid(X_dict["zero_inflation"] @ coef_zero_inflation)

    # negative binomial component
    negbin_loglikelihood = (
        torch.lgamma(y + r)
        - torch.lgamma(r)
        - torch.lgamma(y + 1)
        + r * torch.log(r)
        + y * torch.log(mu)
        - (r + y) * torch.log(r + mu)
    )

    # return the mixture, with an offset to prevent log(0)
    log_likelihood = torch.log(
        pi * (y == 0) + (1 - pi) * torch.exp(negbin_loglikelihood) + 1e-10
    )
    return -torch.sum(log_likelihood)


def zero_inflated_negbin_initializer(X_dict, y, device):
    n_mean_features = X_dict["mean"].shape[1]
    n_dispersion_features = X_dict["dispersion"].shape[1]
    n_zero_inflation_features = X_dict["zero_inflation"].shape[1]
    n_outcomes = y.shape[1]
    return torch.zeros(
        n_mean_features * n_outcomes + n_dispersion_features * n_outcomes + \
        n_zero_inflation_features * n_outcomes, requires_grad=True, device=device
    )


def zero_inflated_negbin_postprocessor(params, X_dict, y):
    n_mean_features = X_dict["mean"].shape[1]
    n_dispersion_features = X_dict["dispersion"].shape[1]
    n_zero_inflation_features = X_dict["zero_inflation"].shape[1]
    n_outcomes = y.shape[1]
    coef_mean = format.to_np(params[:n_mean_features * n_outcomes]).\
        reshape(n_mean_features, n_outcomes)
    coef_dispersion = format.to_np(params[n_mean_features * n_outcomes\
        : n_mean_features * n_outcomes + n_dispersion_features * n_outcomes]).\
        reshape(n_dispersion_features, n_outcomes)
    coef_zero_inflation = format.to_np(params[n_mean_features * n_outcomes \
        + n_dispersion_features * n_outcomes :]).\
        reshape(n_zero_inflation_features, n_outcomes)
    return {"coef_mean": coef_mean, "coef_dispersion": coef_dispersion,\
        "coef_zero_inflation": coef_zero_inflation}


zero_inflated_negbin_regression_array = factory.multiple_formula_regression_factory(
    zero_inflated_negbin_regression_likelihood,
    zero_inflated_negbin_initializer,
    zero_inflated_negbin_postprocessor,
)

###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################


def format_zero_inflated_negbin_parameters(
    parameters: dict, var_names: list, mean_coef_index: 
        list, dispersion_coef_index: list, zero_inflation_coef_index: list
) -> dict:
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=mean_coef_index
    )
    parameters["coef_dispersion"] = pd.DataFrame(
        parameters["coef_dispersion"], columns=var_names, index=dispersion_coef_index
    )
    parameters["coef_zero_inflation"] = pd.DataFrame(
        parameters["coef_zero_inflation"], columns=var_names, index=zero_inflation_coef_index
    )
    return parameters

def format_zero_inflated_negbin_parameters_with_loaders(
    parameters: dict, var_names: list, dls: dict
) -> dict:
    mean_coef_index = dls["mean"].dataset.x_names
    dispersion_coef_index = dls["dispersion"].dataset.x_names
    zero_inflation_coef_index = dls["zero_inflation"].dataset.x_names
    return format_zero_inflated_negbin_parameters(
        parameters, var_names, mean_coef_index, dispersion_coef_index, zero_inflation_coef_index
    )

def standardize_zero_inflated_negbin_formula(formula: Union[str, dict]) -> dict:
    '''
    Convert string formula to dict and validate type.
    If formula is a string, it is the formula for the mean parameter.
    If formula is a dictionary, it is a dictionary of formulas for the mean, dispersion, and zero_inflation parameters.
    '''
    # Convert string formula to dict and validate type
    formula = {'mean': formula, 'dispersion': '~ 1', 'zero_inflation': '~ 1'} \
        if isinstance(formula, str) else formula
    if not isinstance(formula, dict):
        raise ValueError("formula must be a string or a dictionary")
    
    # Define allowed keys and set defaults
    allowed_keys = {'mean', 'dispersion', 'zero_inflation'}
    formula_keys = set(formula.keys())

    # check for required keys and warn about extras
    if not formula_keys & allowed_keys:
        raise ValueError("formula must have at least one of \
                the following keys: mean, dispersion, zero_inflation")
    
    # warn about unused keys
    if extra_keys := formula_keys - allowed_keys:
        warnings.warn(
            f"Invalid formulas in dictionary for zero-inflated \
                negative binomial regression: {extra_keys}",
            UserWarning,
        )
    
    # set default values for missing keys
    formula.update({k: '~ 1' for k in allowed_keys - formula_keys})
    return formula


def zero_inflated_negbin_regression(
    adata: AnnData, formula: Union[str, dict], chunk_size: int = int(1e4), batch_size=512, **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean', 'dispersion', 'zero_inflation'])
    
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )

    parameters = zero_inflated_negbin_regression_array(loaders, **kwargs)
    return format_zero_inflated_negbin_parameters(
        parameters, list(adata.var_names), loaders["mean"].dataset.x_names, 
        loaders["dispersion"].dataset.x_names, loaders["zero_inflation"].dataset.x_names
    )


###############################################################################
## Copula versions for ZINB regression
###############################################################################


def zero_inflated_negbin_uniformizer(parameters, X_dict, y, epsilon=1e-3):
    r, mu, pi = (
        np.exp(X_dict["dispersion"] @ parameters["coef_dispersion"]),
        np.exp(X_dict["mean"] @ parameters["coef_mean"]),
        expit(X_dict["zero_inflation"] @ parameters["coef_zero_inflation"]),
    )
    nb_distn = nbinom(n=r, p=r / (r + mu))
    u1 = pi + (1 - pi) * nb_distn.cdf(y)
    u2 = np.where(y > 0, pi + (1 - pi) * nb_distn.cdf(y-1), 0)
    v = np.random.uniform(size=y.shape)
    return np.clip(v * u1 + (1 - v) * u2, epsilon, 1 - epsilon)


zero_inflated_negbin_copula = gcf.gaussian_copula_factory(
    gcf.gaussian_copula_array_factory(
        zero_inflated_negbin_regression_array, zero_inflated_negbin_uniformizer
    ),
    format_zero_inflated_negbin_parameters_with_loaders,
    param_name=['mean', 'dispersion', 'zero_inflation'],
)



import numpy as np
import pandas as pd
from ..format import format_matrix
from typing import Union

def poisson_predict(parameters: dict, obs: pd.DataFrame, formula: Union[str, dict]):
    if isinstance(formula, dict):
        formula = formula['mean']
    x = format_matrix(obs, formula)
    mu = np.exp(x @ parameters["coef_mean"])
    return {"mean": mu}




from .plot import (
    plot_umap,
    plot_hist,
    compare_means,
    compare_variances,
    compare_standard_deviation,
    compare_umap,
    compare_pca,
)
from .aic_bic import compose_marginal_diagnose, compose_gcopula_diagnose
from .. import estimators as est

__all__ = [
    "bernoulli_gcopula_diagnose",
    "bernoulli_regression_diagnose",
    "compare_means",
    "compare_pca",
    "compare_standard_deviation",
    "compare_umap",
    "compare_variances",
    "gaussian_regression_diagnose",
    "negbin_gcopula_diagnose",
    "negbin_regression_diagnose",
    "plot_hist",
    "plot_pca",
    "plot_umap",
    "poisson_gcopula_diagnose",
    "poisson_regression_diagnose",
    "zinb_gcopula_diagnose",
    "zinb_regression_diagnose",
    "zip_regression_diagnose"
]


###############################################################################
## Methods for calculating marginal/gaussian copula AIC/BIC
###############################################################################

negbin_regression_diagnose = compose_marginal_diagnose(est.negbin.negbin_regression_likelihood,
                                                       allowed_keys=['mean', 'dispersion'])
negbin_gcopula_diagnose = compose_gcopula_diagnose(est.negbin.negbin_regression_likelihood,
                                                   est.negbin.negbin_uniformizer,
                                                   allowed_keys=['mean', 'dispersion'])
poisson_regression_diagnose = compose_marginal_diagnose(est.poisson.poisson_regression_likelihood,
                                                        allowed_keys=['mean'])
poisson_gcopula_diagnose = compose_gcopula_diagnose(est.poisson.poisson_regression_likelihood,
                                                    est.poisson.poisson_uniformizer,
                                                    allowed_keys=['mean'])
bernoulli_regression_diagnose = compose_marginal_diagnose(est.bernoulli.bernoulli_regression_likelihood,
                                                          allowed_keys=['mean'])
bernoulli_gcopula_diagnose = compose_gcopula_diagnose(est.bernoulli.bernoulli_regression_likelihood,
                                                      est.bernoulli.bernoulli_uniformizer,
                                                      allowed_keys=['mean'])
zinb_regression_diagnose = compose_marginal_diagnose(est.zero_inflated_negbin.zero_inflated_negbin_regression_likelihood,
                                                     allowed_keys=['mean', 'dispersion', 'zero_inflation'])
zinb_gcopula_diagnose = compose_gcopula_diagnose(est.zero_inflated_negbin.zero_inflated_negbin_regression_likelihood,
                                                 est.zero_inflated_negbin.zero_inflated_negbin_uniformizer,
                                                 allowed_keys=['mean', 'dispersion', 'zero_inflation'])
zip_regression_diagnose = compose_marginal_diagnose(est.zero_inflated_poisson.zero_inflated_poisson_regression_likelihood,
                                                    allowed_keys=['mean', 'zero_inflation'])
gaussian_regression_diagnose = compose_marginal_diagnose(est.gaussian.gaussian_regression_likelihood,
                                                         allowed_keys=['mean', 'sdev'])
gaussian_gcopula_diagnose = compose_gcopula_diagnose(est.gaussian.gaussian_regression_likelihood, 
                                                     est.gaussian.gaussian_uniformizer,
                                                     allowed_keys=['mean', 'sdev'])



def _filter_kwargs(kwargs: dict, allowed: set) -> dict:
    """Return a new dict containing only keys from kwargs that are in allowed."""
    return {k: v for k, v in kwargs.items() if k in allowed}


# allowed kwargs for common components
DEFAULT_ALLOWED_KWARGS = {
    'trainer': {
        'max_epochs', 'gpus', 'devices', 'accelerator', 'precision',
        'logger', 'callbacks', 'strategy', 'num_nodes', 'limit_train_batches',
        'log_every_n_steps', 'accumulate_grad_batches'
    },
    'data': {
        'chunk_size', 'batch_size', 'shuffle', 'num_workers'
    },
    'optimizer': {
        'lr', 'learning_rate', 'momentum', 'weight_decay', 'eps', 'betas',
        'amsgrad', 'dampening', 'nesterov', 'alpha',
        'T_max', 'eta_min', 'step_size', 'gamma', 'milestones', 'last_epoch',
        'verbose', 'patience', 'threshold', 'cooldown',
        'optimizer_class', 'optimizer', 'scheduler_class', 'scheduler',
        'monitor', 'interval', 'frequency'
    }
}


import numpy as np
import pandas as pd
from ..format import format_matrix
from scipy.special import expit
from typing import Union

def zero_inflated_negbin_predict(parameters: dict, obs: pd.DataFrame, formula: Union[str, dict]):
    if isinstance(formula, str):
        formula = {"mean": formula, "dispersion": "~ 1", "zero_inflation": "~ 1"}
    x_mean = format_matrix(obs, formula["mean"])
    x_dispersion = format_matrix(obs, formula["dispersion"])
    x_zero_inflation = format_matrix(obs, formula["zero_inflation"])
    r, mu, pi = (
        np.exp(x_dispersion @ parameters["coef_dispersion"]),
        np.exp(x_mean @ parameters["coef_mean"]),
        expit(x_zero_inflation @ parameters["coef_zero_inflation"]),
    )
    return {"mean": mu, "dispersion": r, "zero_inflation": pi}



import rich
import rich.table


def shorten_names(features, max_features=5):
    if len(features) > max_features:
        features = features[: (int(max_features - 1))] + ["..."] + features[-1:]
    return f"""[{', '.join(features)}]"""


def print_simulator(margins, copula):
    table = rich.table.Table(
        title="[bold magenta]Simulation Plan[/bold magenta]", title_justify="left"
    )
    table.add_column("formula")
    table.add_column("distribution")
    table.add_column("features")

    i = 1
    for m in margins:
        features, margin = m
        tup = tuple(margin.to_df().iloc[0, :]) + (shorten_names(features),)
        table.add_row(*tup)
        i += 1

    rich.print(table)
    if copula is None:
        rich.print("Marginal models without copula.")

    return ""



from . import gaussian_copula_factory as gcf
from . import glm_factory as factory
from .. import format
from .. import data
from anndata import AnnData
from scipy.stats import nbinom
import numpy as np
import pandas as pd
import torch
from typing import Union

###############################################################################
## Regression functions that operate on numpy arrays
###############################################################################


def negbin_regression_likelihood(params, X_dict, y):    
    n_mean_features = X_dict["mean"].shape[1]
    n_dispersion_features = X_dict["dispersion"].shape[1]
    n_outcomes = y.shape[1]

    # form the mean and dispersion parameters
    coef_mean = params[: n_mean_features * n_outcomes].\
        reshape(n_mean_features, n_outcomes)
    coef_dispersion = params[n_mean_features * n_outcomes :].\
        reshape(n_dispersion_features, n_outcomes)
    r = torch.exp(X_dict["dispersion"] @ coef_dispersion)
    mu = torch.exp(X_dict["mean"] @ coef_mean)

    # compute the negative log likelihood
    log_likelihood = (
        torch.lgamma(y + r)
        - torch.lgamma(r)
        - torch.lgamma(y + 1)
        + r * torch.log(r)
        + y * torch.log(mu)
        - (r + y) * torch.log(r + mu)
    )
    
    return -torch.sum(log_likelihood)


def negbin_initializer(x_dict, y, device):
    n_mean_features = x_dict["mean"].shape[1]
    n_outcomes = y.shape[1]
    n_dispersion_features = x_dict["dispersion"].shape[1]
    return torch.zeros(
        n_mean_features * n_outcomes\
            + n_dispersion_features * n_outcomes, 
        requires_grad=True, device=device
    )


def negbin_postprocessor(params, x_dict, y):
    n_mean_features = x_dict["mean"].shape[1]
    n_outcomes = y.shape[1]
    n_dispersion_features = x_dict["dispersion"].shape[1]
    coef_mean = format.to_np(params[:n_mean_features * n_outcomes]).\
        reshape(n_mean_features, n_outcomes)
    coef_dispersion = format.to_np(params[n_mean_features * n_outcomes:]).\
        reshape(n_dispersion_features, n_outcomes)
    return {"coef_mean": coef_mean, "coef_dispersion": coef_dispersion}


negbin_regression_array = factory.multiple_formula_regression_factory(
    negbin_regression_likelihood, negbin_initializer, negbin_postprocessor
)


###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################

def format_negbin_parameters(
    parameters: dict, var_names: list, mean_coef_index: list, 
    dispersion_coef_index: list
) -> dict:
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=mean_coef_index
    )
    parameters["coef_dispersion"] = pd.DataFrame(
        parameters["coef_dispersion"], columns=var_names, index=dispersion_coef_index
    )
    return parameters

def format_negbin_parameters_with_loaders(
    parameters: dict, var_names: list, dls: dict
) -> dict:
    # Extract the coefficient indices from the dataloaders
    mean_coef_index = dls["mean"].dataset.x_names
    dispersion_coef_index = dls["dispersion"].dataset.x_names
    
    return format_negbin_parameters(parameters, var_names, mean_coef_index, dispersion_coef_index)

def negbin_regression(
    adata: AnnData, formula: Union[str, dict], chunk_size: int = int(1e4), batch_size=512, **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean', 'dispersion'])
    
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )
    parameters = negbin_regression_array(loaders, **kwargs)
    return format_negbin_parameters(
        parameters, list(adata.var_names), loaders["mean"].dataset.x_names, loaders["dispersion"].dataset.x_names
    )

###############################################################################
## Copula versions for negative binomial regression
###############################################################################


def negbin_uniformizer(parameters, X_dict, y, epsilon=1e-3):
    r = np.exp(X_dict["dispersion"] @ parameters["coef_dispersion"])
    mu = np.exp(X_dict["mean"] @ parameters["coef_mean"])
    u1 = nbinom(n=r, p=r / (r + mu)).cdf(y)
    u2 = np.where(y > 0, nbinom(n=r, p=r / (r + mu)).cdf(y - 1), 0)
    v = np.random.uniform(size=y.shape)
    return np.clip(v * u1 + (1 - v) * u2, epsilon, 1 - epsilon)


negbin_copula_array = gcf.gaussian_copula_array_factory(
    negbin_regression_array, negbin_uniformizer
) # should accept a dictionary of dataloaders

negbin_copula = gcf.gaussian_copula_factory(
    negbin_copula_array, format_negbin_parameters_with_loaders, 
    param_name=['mean', 'dispersion']
)



from typing import Union
from copy import deepcopy
import numpy as np


def nullify(params: dict, id: str, mask: Union[np.array, None] = None) -> dict:
    """
    # fit a model
    example_sce = anndata.read_h5ad(save_path)
    x = formulaic.model_matrix("~ bs(pseudotime, degree=2)", example_sce.obs)
    x = x.to_numpy()
    y = example_sce.X.todense()
    params = negative_binomial_copula(x, y)

    # nullify pseudotime for first ten genes
    outcomes = example_sce.var_names[:10]
    mask = anndata_formula_mask(outcomes, ["pseudotime], "~ bs(pseudotime, degree=2)", example_sce)
    null_params = nullify(params, "beta", mask)
    """
    if mask is None:
        mask = np.ones(params[id].shape)

    result = deepcopy(params)
    result[id][mask] = 0
    return result



from .formula import FormulaViewDataset, formula_loader, multiple_formula_loader, standardize_formula
from .group import FormulaGroupViewDataset, formula_group_loader, stack_collate, multiple_formula_group_loader
from .sparse import SparseMatrixDataset, SparseMatrixLoader

__all__ = [
    "FormulaViewDataset",
    "SparseMatrixDataset",
    "SparseMatrixLoader",
    "FormulaGroupViewDataset",
    "formula_loader",
    "formula_group_loader",
    "stack_collate",
    "multiple_formula_loader",
    "multiple_formula_group_loader",
    "standardize_formula"
]


from .scd3_instances import (
    BernoulliCopula,
    GaussianCopula,
    NegBinCopula,
    ZeroInflatedNegBinCopula
)
from .composite import CompositeCopula
from .positive_nonnegative_matrix_factorization import PositiveNMF

__all__ = [
    "BernoulliCopula",
    "CompositeCopula",
    "GaussianCopula",
    "NegBinCopula",
    "PositiveNMF",
    "ZeroInflatedNegBinCopula"
]


from anndata import AnnData
from typing import Dict
from pandas import DataFrame
from abc import abstractmethod

class Simulator:
    """Simulation abstract class"""

    def __init__(self):
        self.parameters = None

    @abstractmethod
    def fit(self, anndata: AnnData, **kwargs) -> None:
        """Fit the simulator"""
        self.template = anndata

    @abstractmethod
    def predict(self, obs: DataFrame=None, **kwargs) -> Dict:
        """Predict from an obs dataframe"""
        pass

    @abstractmethod
    def sample(self, obs: DataFrame=None, **kwargs) -> AnnData:
        """Generate samples."""
        pass


from .composite_regressor import CompositeGLMSimulator
from .glm_simulator import (
    BernoulliCopulaSimulator,
    BernoulliRegressionSimulator,
    NegBinCopulaSimulator,
    NegBinRegressionSimulator,
    PoissonCopulaSimulator,
    PoissonRegressionSimulator,
    GaussianRegressionSimulator,
    GaussianCopulaSimulator,
    ZeroInflatedNegBinCopulaSimulator,
    ZeroInflatedNegBinRegressionSimulator,
    ZeroInflatedPoissonRegressionSimulator,
)
from .pnmf_regression import PNMFRegressionSimulator

__all__ = [
    "BernoulliCopulaSimulator",
    "BernoulliRegressionSimulator",
    "CompositeGLMSimulator",
    "GaussianRegressionSimulator",
    "GaussianCopulaSimulator",
    "NegBinCopulaSimulator",
    "NegBinRegressionSimulator",
    "PNMFRegressionSimulator",
    "PoissonCopulaSimulator",
    "PoissonRegressionSimulator",
    "ZeroInflatedNegBinCopulaSimulator",
    "ZeroInflatedNegBinRegressionSimulator",
    "ZeroInflatedPoissonRegressionSimulator",
]



from ..format import format_matrix
from typing import Union
import numpy as np
import pandas as pd


def zero_inflated_poisson_predict(parameters: dict, obs: pd.DataFrame, formula: Union[str, dict]):
    if isinstance(formula, str):
        formula = {'beta': formula, 'pi': '~ 1'}
    mu, pi = (
        np.exp(format_matrix(obs, formula['mean']) @ parameters["coef_mean"]),
        sigmoid(format_matrix(obs, formula['zero_inflation']) @ parameters["coef_zero_inflation"]),
    )
    return {"mean": mu, "zero_inflation": pi}


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


from .negbin import negbin_regression, negbin_copula
from .gaussian_copula_factory import group_indices
from .poisson import poisson_regression, poisson_copula
from .bernoulli import bernoulli_regression, bernoulli_copula
from .gaussian import gaussian_regression, gaussian_copula
from .zero_inflated_negbin import (
    zero_inflated_negbin_regression,
    zero_inflated_negbin_copula,
)
from .zero_inflated_poisson import zero_inflated_poisson_regression
from .glm_factory import multiple_formula_regression_factory

__all__ = [
    "bernoulli_regression",
    "bernoulli_copula",
    "negbin_copula",
    "negbin_regression",
    "gaussian_regression",
    "gaussian_copula",
    "group_indices",
    "poisson_copula",
    "poisson_regression",
    "zero_inflated_negbin_copula",
    "zero_inflated_negbin_regression",
    "zero_inflated_poisson_regression",
    "multiple_formula_regression_factory",
]



from . import gaussian_copula_factory as gcf
from . import glm_factory as factory
from .. import data
from .. import format
from anndata import AnnData
from scipy.stats import poisson
import numpy as np
import pandas as pd
import torch

###############################################################################
## Regression functions that operate on numpy arrays
###############################################################################


def poisson_regression_likelihood(params, X, y, epsilon=1e-6):
    # get appropriate parameter shape
    n_features = X['mean'].shape[1]
    n_outcomes = y.shape[1]

    # compute the negative log likelihood
    beta = params.reshape(n_features, n_outcomes)
    mu = torch.exp(X['mean'] @ beta)
    log_likelihood = y * torch.log(mu+epsilon) - mu - torch.lgamma(y + 1)
    return -torch.sum(log_likelihood)


def poisson_initializer(x, y, device):
    n_features, n_outcomes = x['mean'].shape[1], y.shape[1]
    return torch.zeros(n_features * n_outcomes, requires_grad=True, device=device)


def poisson_postprocessor(params, x, y):
    coef_mean = format.to_np(params).reshape(x['mean'].shape[1], y.shape[1])
    return {"coef_mean": coef_mean}


poisson_regression_array = factory.multiple_formula_regression_factory(
    poisson_regression_likelihood, poisson_initializer, poisson_postprocessor
)

###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################


def format_poisson_parameters(
    parameters: dict, var_names: list, coef_index: list
) -> dict:
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=coef_index
    )
    return parameters


def poisson_regression(
    adata: AnnData,
    formula: str,
    chunk_size: int = int(1e4),
    batch_size: int = 512,
    **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean'])
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )

    parameters = poisson_regression_array(loaders, **kwargs)
    return format_poisson_parameters(
        parameters, list(adata.var_names), loaders["mean"].dataset.x_names
    )


###############################################################################
## Copula versions for poisson regression
###############################################################################


def poisson_uniformizer(parameters, x, y, epsilon=1e-3):
    mu = np.exp(x['mean'] @ parameters["coef_mean"])
    u1 = poisson(mu).cdf(y)
    u2 = np.where(y > 0, poisson(mu).cdf(y - 1), 0)
    v = np.random.uniform(size=y.shape)
    return np.clip(v * u1 + (1 - v) * u2, epsilon, 1 - epsilon)

def format_poisson_parameters_with_loaders(parameters: dict, var_names: list, dls: dict) -> dict:
    beta_coef_index = dls["mean"].dataset.x_names
    
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=beta_coef_index
    )
    return parameters

poisson_copula_array = gcf.gaussian_copula_array_factory(
    poisson_regression_array, poisson_uniformizer
) 

poisson_copula = gcf.gaussian_copula_factory(
    poisson_copula_array, format_poisson_parameters_with_loaders, ['mean']
)



from .kwargs import DEFAULT_ALLOWED_KWARGS, _filter_kwargs
from .loader import adata_loader
from anndata import AnnData
from typing import Union, Dict, Optional, Tuple
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn


class Marginal:
    def __init__(self, formula: Union[Dict, str]):
        self.formula = formula
        self.feature_dims = None
        self.loader = None
        self.n_outcomes = None
        self.predict = None
        self.predictor_names = None
        self.parameters = None

    def setup_data(self, adata: AnnData, batch_size: int = 1024, **kwargs):
        """Set up the dataloader for the AnnData object."""
        # keep a reference to the AnnData for later use (e.g., var_names)
        self.adata = adata
        self.loader = adata_loader(adata, self.formula, batch_size=batch_size, **kwargs)
        X_batch, obs_batch = next(iter(self.loader))
        self.n_outcomes = X_batch.shape[1]
        self.feature_dims = {k: v.shape[1] for k, v in obs_batch.items()}
        self.predictor_names = self.loader.dataset.predictor_names

    def fit(self, **kwargs):
        """Fit the marginal predictor"""
        if self.predict is None:
            self.setup_optimizer(**kwargs)
        trainer_kwargs = _filter_kwargs(kwargs, DEFAULT_ALLOWED_KWARGS['trainer'])
        trainer = pl.Trainer(**trainer_kwargs)
        trainer.fit(self.predict, train_dataloaders=self.loader)
        self.parameters = self.format_parameters()

    def setup_optimizer(self, **kwargs):
        raise NotImplementedError

    def likelihood(self, batch: Tuple[torch.Tensor, Dict[str, torch.Tensor]]):
        """Compute the (negative) log-likelihood or loss for a batch.
        """
        raise NotImplementedError

    def invert(self, u: torch.Tensor, x: Dict[str, torch.Tensor]):
        """Invert pseudoobservations."""
        raise NotImplementedError

    def uniformize(self, y: torch.Tensor, x: Dict[str, torch.Tensor]):
        """Uniformize using learned CDF.
        """
        raise NotImplementedError

    def format_parameters(self):
        """Convert fitted coefficient tensors into pandas DataFrames.

        Returns:
            dict: mapping from parameter name -> pandas.DataFrame with rows
                corresponding to predictor column names (from
                `self.predictor_names[param]`) and columns corresponding to
                `self.adata.var_names` (gene names). The values are moved to
                CPU and converted to numpy floats.
        """
        var_names = list(self.adata.var_names)

        dfs = {}
        for param, tensor in self.predict.coefs.items():
            coef_np = tensor.detach().cpu().numpy()
            row_names = list(self.predictor_names[param])
            dfs[param] = pd.DataFrame(coef_np, index=row_names, columns=var_names)
        return dfs

    def num_params(self):
        """Return the number of parameters."""
        if self.predict is None:
            return 0
        return sum(p.numel() for p in self.predict.parameters() if p.requires_grad)


class GLMPredictor(pl.LightningModule):
    """GLM-style predictor with arbitrary named parameters.

    Args:
        n_outcomes: number of model outputs (e.g. genes)
        feature_dims: mapping from param name -> number of covariate features
        link_fns: optional mapping from param name -> callable(link) applied to linear predictor

    The module will create one coefficient matrix per named parameter with shape
    (n_features_for_param, n_outcomes) and expose them as Parameters under
    `self.coefs[param_name]`.
    """
    def __init__(
        self,
        n_outcomes: int,
        feature_dims: Dict[str, int],
        link_fns: Dict[str, callable] = None,
        loss_fn: Optional[callable] = None,
        optimizer_class: Optional[callable] = torch.optim.Adam,
        optimizer_kwargs: Optional[Dict] = None,
    ):
        super().__init__()
        self.n_outcomes = int(n_outcomes)
        self.feature_dims = dict(feature_dims)
        self.param_names = list(self.feature_dims.keys())

        # create default link functions and parameter matrices
        self.link_fns = link_fns or {k: torch.exp for k in self.param_names}
        self.coefs = nn.ParameterDict()
        for key, dim in self.feature_dims.items():
            self.coefs[key] = nn.Parameter(torch.zeros(dim, self.n_outcomes))

        # optimization parameters
        self.reset_parameters()
        self.loss_fn = loss_fn
        self.optimizer_class = optimizer_class
        self.optimizer_kwargs = optimizer_kwargs

    def reset_parameters(self):
        for p in self.coefs.values():
            nn.init.normal_(p, mean=0.0, std=1e-2)

    def forward(self, obs_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        out = {}
        for name in self.param_names:
            x_beta = obs_dict[name] @ self.coefs[name]
            link = self.link_fns.get(name, torch.exp)
            out[name] = link(x_beta)
        return out

    def training_step(self, batch):
        loss = self.loss_fn(batch)
        self.log('train_loss', loss, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    def configure_optimizers(self, **kwargs):
        optimizer_kwargs = _filter_kwargs(kwargs, DEFAULT_ALLOWED_KWARGS['optimizer'])
        return self.optimizer_class(self.parameters(), **optimizer_kwargs)


from .nullify import nullify
from .amplify import amplify
from .substitute import substitute
from .split import split_glm
from .mask import data_frame_mask

__all__ = ["amplify", "nullify", "data_frame_mask", "substitute", "split_glm"]



from anndata import AnnData
from formulaic import model_matrix
import pandas as pd
import scipy.sparse

def to_np(x):
    return x.detach().cpu().numpy()

def format_input_anndata(adata: AnnData) -> AnnData:
    result = adata.copy()
    if isinstance(result.X, scipy.sparse._csc.csc_matrix):
        result.X = result.X.todense()
    return result

def format_matrix(obs: pd.DataFrame, formula: str):
    if formula is not None:
        x = model_matrix(formula, pd.DataFrame(obs))
    else:
        x = obs
    return x


from .scd3 import SCD3Simulator
from .negbin import NegBin
from .zero_inflated_negbin import ZeroInflatedNegBin
from .gaussian import Gaussian
from .standard_covariance import StandardCovariance
from typing import Optional


class NegBinCopula(SCD3Simulator):
    def __init__(self,
                 mean_formula: Optional[str] = None,
                 dispersion_formula: Optional[str] = None,
                 copula_formula: Optional[str] = None) -> None:
        marginal = NegBin({"mean": mean_formula, "dispersion": dispersion_formula})
        covariance = StandardCovariance(copula_formula)
        super().__init__(marginal, covariance)


class ZeroInflatedNegBinCopula(SCD3Simulator):
    def __init__(self,
                 mean_formula: Optional[str] = None,
                 dispersion_formula: Optional[str] = None,
                 zero_inflation_formula: Optional[str] = None,
                 copula_formula: Optional[str] = None) -> None:
        marginal = ZeroInflatedNegBin({
            "mean": mean_formula,
            "dispersion": dispersion_formula,
            "zero_inflation_formula": zero_inflation_formula
        })
        covariance = StandardCovariance(copula_formula)
        super().__init__(marginal, covariance)


class BernoulliCopula(SCD3Simulator):
    def __init__(self,
                 mean_formula: Optional[str] = None,
                 copula_formula: Optional[str] = None) -> None:
        marginal = NegBin({"mean": mean_formula})
        covariance = StandardCovariance(copula_formula)
        super().__init__(marginal, covariance)


class GaussianCopula(SCD3Simulator):
    def __init__(self,
                 mean_formula: Optional[str] = None,
                 sdev_formula: Optional[str] = None,
                 copula_formula: Optional[str] = None) -> None:
        marginal = Gaussian({"mean": mean_formula, "sdev": sdev_formula})
        covariance = StandardCovariance(copula_formula)
        super().__init__(marginal, covariance)


from .negbin import negbin_sample, negbin_copula_sample
from .poisson import poisson_sample, poisson_copula_sample
from .bernoulli import bernoulli_sample, bernoulli_copula_sample
from .gaussian import gaussian_regression_sample, gaussian_copula_sample
from .zero_inflated_negbin import (
    zero_inflated_negbin_sample,
    zero_inflated_negbin_copula_sample,
)
from .zero_inflated_poisson import zero_inflated_poisson_sample

__all__ = [
    "negbin_sample",
    "negbin_copula_sample",
    "poisson_sample",
    "poisson_copula_sample",
    "bernoulli_sample",
    "bernoulli_copula_sample",
    "gaussian_regression_sample",
    "gaussian_copula_sample",
    "zero_inflated_negbin_sample",
    "zero_inflated_negbin_copula_sample",
    "zero_inflated_poisson_sample",
]



import pandas as pd
from . import gaussian_copula_factory as gcf
from . import glm_factory as factory
from .. import data
from .. import format
from . import poisson as poi
from anndata import AnnData
from scipy.stats import bernoulli
import numpy as np
import torch
from typing import Union

###############################################################################
## Regression functions that operate on numpy arrays
###############################################################################


def bernoulli_regression_likelihood(params, X_dict, y):
    # get appropriate parameter shape
    n_features = X_dict["mean"].shape[1]
    n_outcomes = y.shape[1]

    # compute the negative log likelihood
    coef_mean = params.reshape(n_features, n_outcomes)
    theta = torch.sigmoid(X_dict["mean"] @ coef_mean)
    log_likelihood = y * torch.log(theta) + (1 - y) * torch.log(1 - theta)
    return -torch.sum(log_likelihood)

def bernoulli_initializer(X_dict, y, device):
    n_features = X_dict["mean"].shape[1]
    n_outcomes = y.shape[1]
    return torch.zeros(n_features * n_outcomes, requires_grad=True, device=device)

def bernoulli_postprocessor(params, X_dict, y):
    coef_mean = format.to_np(params).reshape(X_dict["mean"].shape[1], y.shape[1])
    return {"coef_mean": coef_mean}

bernoulli_regression_array = factory.multiple_formula_regression_factory(
    bernoulli_regression_likelihood, bernoulli_initializer, bernoulli_postprocessor
)

###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################


def bernoulli_regression(
    adata: AnnData, formula: Union[str, dict], chunk_size: int = int(1e4), batch_size=512, **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean'])
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )
    parameters = bernoulli_regression_array(loaders, **kwargs)
    return poi.format_poisson_parameters(
        parameters, list(adata.var_names), list(loaders["mean"].dataset.x_names)
    )


###############################################################################
## Copula versions for bernoulli regression
###############################################################################


def bernoulli_uniformizer(parameters, X_dict, y, epsilon=1e-3, random_seed=42):
    np.random.seed(random_seed)
    theta = torch.sigmoid(torch.from_numpy(X_dict["mean"] @ parameters["coef_mean"])).numpy()
    u1 =  bernoulli(theta).cdf(y)
    u2 = np.where(y > 0,  bernoulli(theta).cdf(y - 1), 0)
    v = np.random.uniform(size=y.shape)
    return np.clip(v * u1 + (1 - v) * u2, epsilon, 1 - epsilon)

def format_bernoulli_parameters_with_loaders(parameters: dict, var_names: list, dls: dict) -> dict:
    coef_mean_index = dls["mean"].dataset.x_names
    
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=coef_mean_index
    )
    return parameters

bernoulli_copula = gcf.gaussian_copula_factory(
    gcf.gaussian_copula_array_factory(bernoulli_regression_array, bernoulli_uniformizer),
    format_bernoulli_parameters_with_loaders,
    param_name=['mean']
)



from scipy.stats import poisson, bernoulli
from . import glm_factory as glm
import numpy as np


def zero_inflated_poisson_sample_array(local_parameters: dict) -> np.array:
    mu, pi = (
        local_parameters["mean"],
        local_parameters["zero_inflation"],
    )
    return poisson(mu).rvs() * bernoulli(1 - pi).rvs()


zero_inflated_poisson_sample = glm.glm_sample_factory(
    zero_inflated_poisson_sample_array
)



from .copula import Copula
from .formula import standardize_formula
from .kwargs import DEFAULT_ALLOWED_KWARGS, _filter_kwargs
from anndata import AnnData
from scipy.stats import norm, multivariate_normal
from tqdm import tqdm
from typing import Dict, Union, Callable, Tuple
import numpy as np
import pandas as pd
import torch


class StandardCovariance(Copula):
    def __init__(self, formula: str = "~ 1"):
        formula = standardize_formula(formula, allowed_keys=['group'])
        super().__init__(formula)
        self.groups = None


    def setup_data(self, adata: AnnData, marginal_formula: Dict[str, str], **kwargs):
        data_kwargs = _filter_kwargs(kwargs, DEFAULT_ALLOWED_KWARGS['data'])
        super().setup_data(adata, marginal_formula, **data_kwargs)
        _, obs_batch = next(iter(self.loader))
        obs_batch_group = obs_batch.get("group")

        # fill in group indexing variables
        self.groups = self.loader.dataset.predictor_names["group"]
        self.n_groups = len(self.groups)
        self.group_col = {g: i for i, g in enumerate(self.groups)}

        # check that obs_batch is a binary grouping matrix
        unique_vals = torch.unique(obs_batch_group)
        if (not torch.all((unique_vals == 0) | (unique_vals == 1)).item()):
            raise ValueError("Only categorical groups are currently supported in copula covariance estimation.")

    def fit(self, uniformizer: Callable, **kwargs):
        sums = {g: np.zeros(self.n_outcomes) for g in self.groups}
        second_moments = {g: np.eye(self.n_outcomes) for g in self.groups}
        Ng = {g: 0 for g in self.groups}

        for y, x_dict in tqdm(self.loader, desc="Estimating copula covariance"):
            memberships = x_dict.get("group").numpy()
            u = uniformizer(y, x_dict)

            for g in self.groups:
                ix = np.where(memberships[:, self.group_col[g]] == 1)
                z = norm().ppf(u[ix])
                second_moments[g] += z.T @ z
                sums[g] += z.sum(axis=0)
                Ng[g] += len(ix[0])

        covariances = {}
        for g in self.groups:
            mean = sums[g] / Ng[g]
            covariances[g] = second_moments[g] / Ng[g] - np.outer(mean, mean)

        if len(self.groups) == 1:
            covariances = list(covariances.values())[0]
        self.parameters = self.format_parameters(covariances)

    def format_parameters(self, covariances: Union[Dict, np.array]):
        var_names = self.adata.var_names
        def to_df(mat):
            return pd.DataFrame(mat, index=var_names, columns=var_names)

        if isinstance(covariances, dict):
            formatted = {}
            for k, v in covariances.items():
                formatted[k] = to_df(v)
            covariances = formatted
            return covariances

        if isinstance(covariances, (np.ndarray, list, tuple)):
            covariances = to_df(covariances)
        return covariances

    def pseudo_obs(self, x_dict: Dict):
        # convert one-hot encoding memberships to a map
        #      {"group1": [indices of group 1], "group2": [indices of group 2]}
        memberships = x_dict.get("group").numpy()
        group_ix = {g: np.where(memberships[:, self.group_col[g] == 1])[0] for g in self.groups}

        # initialize the result
        u = np.zeros((len(memberships), self.n_outcomes))
        parameters = self.parameters
        if type(parameters) is not dict:
            parameters = {group: parameters}

        # loop over groups and sample each part in turn
        for group, sigma in parameters.items():
            z = np.random.multivariate_normal(
                mean=np.zeros(self.n_outcomes),
                cov=sigma,
                size=len(group_ix[group])
            )
            normal_distn = norm(0, np.diag(sigma) ** 0.5)
            u[group_ix[group]] = normal_distn.cdf(z)
        return u

    def likelihood(self, uniformizer: Callable, batch: Tuple[torch.Tensor, Dict[str, torch.Tensor]]):
        # uniformize the observations
        y, x_dict = batch
        u = uniformizer(y, x_dict)
        z = norm().ppf(u)

        # same group manipulation as for pseudobs
        parameters = self.parameters
        if type(parameters) is not dict:
            parameters = {group: parameters}

        memberships = x_dict.get("group").numpy()
        group_ix = {g: np.where(memberships[:, self.group_col[g] == 1])[0] for g in self.groups}
        ll = np.zeros(len(z))
        for group, sigma in parameters.items():
            ix = group_ix[group]
            if len(ix) > 0:
                copula_ll = multivariate_normal.logpdf(z[ix], np.zeros(sigma.shape[0]), sigma)
                ll[ix] = copula_ll - norm.logpdf(z[ix]).sum(axis=1)
        return ll

    def num_params(self, **kwargs):
        S = self.parameters
        per_group = [(np.sum(S[g].values != 0) - S[g].shape[0]) / 2 for g in self.groups]
        return sum(per_group)


from ..data import stack_collate, multiple_formula_group_loader
from .. import data
from anndata import AnnData
from collections.abc import Callable
from typing import Union
from scipy.stats import norm
from torch.utils.data import DataLoader
import numpy as np
import pandas as pd

###############################################################################
## General copula factory functions
###############################################################################


def gaussian_copula_array_factory(marginal_model: Callable, uniformizer: Callable):
    def copula_fun(loaders: dict[str, DataLoader], lr: float = 0.1, epochs: int = 40, **kwargs):
        # for the marginal model, ignore the groupings
        # Strip all dataloaders and create a dictionary to pass to marginal_model
        formula_loaders = {}
        for key in loaders.keys():
            formula_loaders[key] = strip_dataloader(loaders[key], pop="Stack" in type(loaders[key].dataset).__name__)
        
        # Call marginal_model with the dictionary of stripped dataloaders
        parameters = marginal_model(formula_loaders, lr=lr, epochs=epochs, **kwargs)

        # estimate covariance, allowing for different groups
        parameters["covariance"] = copula_covariance(parameters, loaders, uniformizer)
        return parameters

    return copula_fun


def gaussian_copula_factory(copula_array_fun: Callable, 
                            parameter_formatter: Callable, 
                            param_name: list = None):
    def copula_fun(
        adata: AnnData,
        formula: Union[str, dict] = "~ 1",
        grouping_var: str = None,
        chunk_size: int = int(1e4),
        batch_size: int = 512,
        **kwargs
    ) -> dict:  
        
        if param_name is not None:
            formula = data.standardize_formula(formula, param_name)
        
        dls = multiple_formula_group_loader(
            adata,
            formula,
            grouping_var,
            chunk_size=chunk_size,
            batch_size=batch_size,
        ) # returns a dictionary of dataloaders
        parameters = copula_array_fun(dls, **kwargs)
        
        # Pass the full dls to parameter_formatter so it can extract what it needs
        parameters = parameter_formatter(
            parameters, adata.var_names, dls
        )
        parameters["covariance"] = format_copula_parameters(parameters, adata.var_names)
        return parameters

    return copula_fun


def copula_covariance(parameters: dict, loaders: dict[str, DataLoader], uniformizer: Callable):
    first_loader = next(iter(loaders.values()))
    D = next(iter(first_loader))[1].shape[1] #dimension of y
    groups = first_loader.dataset.groups # a list of strings of group names
    sums = {g: np.zeros(D) for g in groups}
    second_moments = {g: np.eye(D) for g in groups}
    Ng = {g: 0 for g in groups}
    keys = list(loaders.keys())
    loaders = list(loaders.values())
    num_keys = len(keys)
    
    for batches in zip(*loaders):
        x_batch_dict = {
            keys[i]: batches[i][0].cpu().numpy() for i in range(num_keys)
        }
        y_batch = batches[0][1].cpu().numpy()
        memberships = batches[0][2] # should be identical for all keys
        
        u = uniformizer(parameters, x_batch_dict, y_batch)
        for g in groups:
            ix = np.where(np.array(memberships) == g)
            z = norm().ppf(u[ix])
            second_moments[g] += z.T @ z
            sums[g] += z.sum(axis=0)
            Ng[g] += len(ix[0])

    result = {}
    for g in groups:
        mean = sums[g] / Ng[g]
        result[g] = second_moments[g] / Ng[g] - np.outer(mean, mean)

    if len(groups) == 1:
        return list(result.values())[0] 
    return result 


###############################################################################
## Helpers to prepare and postprocess copula parameters
###############################################################################


def group_indices(grouping_var: str, obs: pd.DataFrame) -> dict:
    """
    Returns a dictionary of group indices for each group in the grouping variable.
    """
    if grouping_var is None:
        grouping_var = "_copula_group"
        if "copula_group" not in obs.columns:
            obs["_copula_group"] = pd.Categorical(["shared_group"] * len(obs))
    result = {}

    for group in list(obs[grouping_var].dtype.categories):
        result[group] = np.where(obs[grouping_var].values == group)[0]
    return result


def clip(u: np.array, min: float = 1e-5, max: float = 1 - 1e-5) -> np.array:
    u[u < min] = min
    u[u > max] = max
    return u


def format_copula_parameters(parameters: dict, var_names: list):
    covariance = parameters["covariance"]
    if type(covariance) is not dict:
        covariance = pd.DataFrame(
            parameters["covariance"], columns=list(var_names), index=list(var_names)
        )
    else:
        for group in covariance.keys():
            covariance[group] = pd.DataFrame(
                parameters["covariance"][group],
                columns=list(var_names),
                index=list(var_names),
            )
    return covariance


def strip_dataloader(dataloader, pop=False):
    return DataLoader(
        dataset=dataloader.dataset,
        batch_sampler=dataloader.batch_sampler,
        collate_fn=stack_collate(pop=pop, groups=False),
    )
    


from tqdm import tqdm
from torch.utils.data import DataLoader
import torch


def glm_regression_factory(likelihood, initializer, postprocessor) -> dict:
    def estimator(
        dataloader: DataLoader,
        lr: float = 0.1,
        epochs: int = 40,
    ):
        device = check_device()
        x, y = next(iter(dataloader))
        params = initializer(x, y, device)
        optimizer = torch.optim.Adam([params], lr=lr)

        for epoch in range(epochs):
            for x_batch, y_batch in (pbar := tqdm(dataloader, desc=f"Epoch {epoch + 1}/{epochs}", leave=False)):
                optimizer.zero_grad()
                loss = likelihood(params, x_batch, y_batch)
                loss.backward()
                optimizer.step()
                pbar.set_postfix_str(f"loss: {loss.item()}")


        return postprocessor(params, x.shape[1], y.shape[1])

    return estimator

def multiple_formula_regression_factory(likelihood, initializer, postprocessor) -> dict:
    def estimator(
        dataloaders: dict[str, DataLoader],
        lr: float = 0.1,
        epochs: int = 40,
    ):  
        device = check_device()
        x_dict = {}
        y_dict = {}
        for key in dataloaders.keys():
            x_dict[key], y_dict[key] = next(iter(dataloaders[key]))
        # check if all ys are the same
        y_ref = y_dict[list(dataloaders.keys())[0]]
        for key in dataloaders.keys():
            if not torch.equal(y_dict[key], y_ref):
                raise ValueError(f"Ys are not the same for {key}")
        params = initializer(x_dict, y_ref, device) # x is a dictionary of tensors, y is a tensor
        optimizer = torch.optim.Adam([params], lr=lr)
        
        keys = list(dataloaders.keys())
        loaders = list(dataloaders.values())
        
        for epoch in range(epochs):
            num_keys = len(keys)
            for batches in (pbar := tqdm(zip(*loaders), desc=f"Epoch {epoch + 1}/{epochs}", leave=False)):
                x_batch_dict = {
                    keys[i]: batches[i][0].to(device) for i in range(num_keys)
                }
                y_batch = batches[0][1].to(device)
                optimizer.zero_grad()
                loss = likelihood(params, x_batch_dict, y_batch) 
                loss.backward()
                optimizer.step()
                pbar.set_postfix_str(f"loss: {loss.item()}")

        return postprocessor(params, x_dict, y_ref)

    return estimator


def check_device():
    return torch.device(
        "cuda"
        if torch.cuda.is_available()
        else "mps" if torch.backends.mps.is_available() else "cpu"
    )



from .kwargs import DEFAULT_ALLOWED_KWARGS, _filter_kwargs
from anndata import AnnData
from formulaic import model_matrix
from torch.utils.data import Dataset, DataLoader
from typing import Dict
import numpy as np
import pandas as pd
import torch

class AnnDataDataset(Dataset):
    """Simple PyTorch Dataset for AnnData objects.

    Supports optional chunked loading for backed AnnData objects. When
    `chunk_size` is provided, the dataset will load contiguous slices
    of rows (of size `chunk_size`) into memory once and serve individual
    rows from that cached chunk. This avoids calling `to_memory()` on
    a per-row basis which is expensive for large backed files.
    """
    def __init__(self, adata: AnnData, formula: Dict[str, str], chunk_size: int):
        self.adata = adata
        self.formula = formula
        self.chunk_size = chunk_size

        # keeping track of covariate-related information
        self.obs_levels = categories(self.adata.obs)
        self.obs_matrices = {}
        self.predictor_names = None

        # Internal cache for the currently loaded chunk
        self._chunk: AnnData | None = None
        self._chunk_start = 0

    def __len__(self):
        return len(self.adata)

    def __getitem__(self, idx):
        """Returns (X, obs) for the given index.

        If `chunk_size` was specified the dataset will load a chunk
        containing `idx` into memory (if not already cached) and
        index into that chunk.
        """
        self._ensure_chunk_loaded(idx)
        local_idx = idx - self._chunk_start
        adata_slice = self._chunk[local_idx]

        # Get X data, accounting for potential sparse matrices
        X = adata_slice.X
        if hasattr(X, 'toarray'):
            X = X.toarray()

        # Get obs data
        obs_dict = {}
        for key in self.formula.keys():
            mat = self.obs_matrices.get(key)
            obs_dict[key] = to_tensor(mat.values[local_idx: local_idx + 1])
        return to_tensor(X), obs_dict

    def _ensure_chunk_loaded(self, idx: int) -> None:
        """Load the chunk that contains `idx` into the internal cache."""
        start = (idx // self.chunk_size) * self.chunk_size
        end = min(start + self.chunk_size, len(self.adata))

        if (self._chunk is None) or not (self._chunk_start <= idx < self._chunk_start + len(self._chunk)):
            # load the next chunk into memory
            chunk = self.adata[start:end]
            if getattr(chunk, 'isbacked', False):
                chunk = chunk.to_memory()
            self._chunk = chunk
            self._chunk_start = start

            # Compute model matrices for this chunk's `obs` so we don't need
            # to keep the full obs data model matrices in memory.
            obs_coded_chunk = code_levels(self._chunk.obs.copy(), self.obs_levels)
            self.obs_matrices = {}
            for key, f in self.formula.items():
                self.obs_matrices[key] = model_matrix(f, obs_coded_chunk)

            # Capture predictor (column) names from the model matrices once.
            if self.predictor_names is None:
                self.predictor_names = {k: list(v.columns) for k, v in self.obs_matrices.items()}


def adata_loader(adata: AnnData,
                 formula: Dict[str, str],
                 chunk_size: int = None,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 num_workers: int = 0,
                 **kwargs) -> DataLoader:
    """
    Create a DataLoader from AnnData that returns batches of (X, obs).
    """
    data_kwargs = _filter_kwargs(kwargs, DEFAULT_ALLOWED_KWARGS['data'])
    if chunk_size is None:
        if getattr(adata, 'isbacked', False):
            chunk_size = 5000
        else:
            chunk_size = len(adata)

    dataset = AnnDataDataset(adata, formula, chunk_size)
    return DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=num_workers,
        collate_fn=dict_collate_fn,
        **data_kwargs
    )

def obs_loader(obs: pd.DataFrame, marginal_formula, **kwargs):
        adata = AnnData(X=np.zeros((len(obs), 1)), obs=obs)
        return adata_loader(
            adata,
            marginal_formula,
            **kwargs
        )

################################################################################
## Helper functions
################################################################################

def dict_collate_fn(batch):
    """
    Custom collate function for handling dictionary obs tensors.
    """
    X_batch = torch.stack([item[0] for item in batch])
    obs_batch = [item[1] for item in batch]

    obs_dict = {}
    for key in obs_batch[0].keys():
        obs_dict[key] = torch.stack([obs[key] for obs in obs_batch])
    return X_batch, obs_dict

def to_tensor(X):
    # If the tensor is 2D with second dim == 1, squeeze only the first
    # dim when appropriate (e.g. converting a single-row X to 1D samples)
    t = torch.tensor(X, dtype=torch.float32)
    if t.dim() == 2 and t.size(1) == 1:
        if t.size(0) == 1:
            return t.view(1)
        return t
    return t.squeeze()

def categories(obs):
    levels = {}
    for k in obs.columns:
        obs_type = str(obs[k].dtype)
        if obs_type in ["category", "object"]:
            levels[k] = obs[k].unique()
    return levels


def code_levels(obs, categories):
    for k in obs.columns:
        if str(obs[k].dtype) == "category":
            obs[k] = obs[k].astype(pd.CategoricalDtype(categories[k]))
    return obs

###############################################################################
## Misc. Helper functions
###############################################################################

def _to_numpy(*tensors):
    """Convenience helper: detach, move to CPU, and convert tensors to numpy arrays."""
    return tuple(t.detach().cpu().numpy() for t in tensors)



from .formula import standardize_formula
from .marginal import GLMPredictor, Marginal
from .loader import _to_numpy
from typing import Union, Dict, Optional
import torch
import numpy as np
from scipy.stats import norm

class Gaussian(Marginal):
    """Gaussian marginal estimator"""
    def __init__(self, formula: Union[Dict, str]):
        formula = standardize_formula(formula, allowed_keys=['mean', 'sdev'])
        super().__init__(formula)

    def setup_optimizer(
            self,
            optimizer_class: Optional[callable] = torch.optim.Adam,
            **optimizer_kwargs,
    ):
        if self.loader is None:
            raise RuntimeError("self.loader is not set (call setup_data first)")

        nll = lambda batch: -self.likelihood(batch).sum()
        link_fns = {"mean": lambda x: x}
        self.predict = GLMPredictor(
            n_outcomes=self.n_outcomes,
            feature_dims=self.feature_dims,
            link_fns=link_fns,
            loss_fn=nll,
            optimizer_class=optimizer_class,
            optimizer_kwargs=optimizer_kwargs
        )

    def likelihood(self, batch):
        """Compute the negative log-likelihood"""
        y, x = batch
        params = self.predict(x)
        mu = params.get("mean")
        sigma = params.get("sdev")

        # log likelihood for Gaussian
        log_likelihood = -0.5 * (torch.log(2 * torch.pi * sigma ** 2) + ((y - mu) ** 2) / (sigma ** 2))
        return log_likelihood

    def invert(self, u: torch.Tensor, x: Dict[str, torch.Tensor]):
        """Invert pseudoobservations."""
        mu, sdev, u = self._local_params(x, u)
        y = norm(loc=mu, scale=sdev).ppf(u)
        return torch.from_numpy(y).float()

    def uniformize(self, y: torch.Tensor, x: Dict[str, torch.Tensor], epsilon=1e-6):
        """Return uniformized pseudo-observations for counts y given covariates x."""
        # cdf values using scipy's parameterization
        mu, sdev, y = self._local_params(x, y)
        u = norm.cdf(y, loc=mu, scale=sdev)
        u = np.clip(u, epsilon, 1 - epsilon)
        return torch.from_numpy(u).float()

    def _local_params(self, x, y=None):
        params = self.predict(x)
        mu = params.get('mean')
        sdev = params.get('sdev')
        if y is None:
            return _to_numpy(mu, sdev)
        return _to_numpy(mu, sdev, y)



from .loader import obs_loader
from .scd3 import SCD3Simulator
from .standard_covariance import StandardCovariance
from anndata import AnnData
from typing import Dict, Optional, List
import numpy as np
import torch

class CompositeCopula(SCD3Simulator):
    def __init__(self, marginals: List,
                 copula_formula: Optional[str] = None) -> None:
        self.marginals = marginals
        self.copula = StandardCovariance(copula_formula)
        self.template = None
        self.parameters = None
        self.merged_formula = None

    def fit(
        self,
        adata: AnnData,
        **kwargs):
        """Fit the simulator"""
        self.template = adata
        merged_formula = {}

        # fit each marginal model
        for m in range(len(self.marginals)):
            self.marginals[m][1].setup_data(adata[:, self.marginals[m][0]], **kwargs)
            self.marginals[m][1].setup_optimizer(**kwargs)
            self.marginals[m][1].fit(**kwargs)

            # prepare formula for copula loader
            f = self.marginals[m][1].formula
            prefixed_f = {f"group{m}_{k}": v for k, v in f.items()}
            merged_formula = merged_formula | prefixed_f

        # copula simulator
        self.merged_formula = merged_formula
        self.copula.setup_data(adata, merged_formula, **kwargs)
        self.copula.fit(self.merged_uniformize, **kwargs)
        self.parameters = {
            "marginal": [m[1].parameters for m in self.marginals],
            "copula": self.copula.parameters
        }

    def merged_uniformize(self, y: torch.Tensor, x: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Produce a merged uniformized matrix for all marginals.

        Delegates to each marginal's `uniformize` method and places the
        result into the columns of a full matrix according to the variable
        selection given in `self.marginals[m][0]`.
        """
        y_np = y.detach().cpu().numpy()
        u = np.empty_like(y_np, dtype=float)

        for m in range(len(self.marginals)):
            sel = self.marginals[m][0]
            ix = _var_indices(sel, self.template)

            # remove the `group{m}_` prefix we used to distinguish the marginals
            prefix = f"group{m}_"
            cur_x = {k.removeprefix(prefix): v if k.startswith(prefix) else v for k, v in x.items()}

            # slice the subset of y for this marginal and call its uniformize
            y_sub = torch.from_numpy(y_np[:, ix])
            u[:, ix] = self.marginals[m][1].uniformize(y_sub, cur_x)
        return torch.from_numpy(u)

    def predict(self, obs=None, batch_size: int = 1000, **kwargs):
        """Predict from an obs dataframe"""
        # prepare an internal data loader for this obs
        if obs is None:
            obs = self.template.obs
        loader = obs_loader(
            obs,
            self.merged_formula,
            batch_size=batch_size,
            **kwargs
        )

        # prepare per-marginal collectors
        n_marginals = len(self.marginals)
        local_pred = [[] for _ in range(n_marginals)]

        # for each batch, call each marginal's predict on its subset of x
        for _, x_dict in loader:
            for m in range(n_marginals):
                prefix = f"group{m}_"
                # build cur_x where prefixed keys are unprefixed for the marginal
                cur_x = {k.removeprefix(prefix): v for k, v in x_dict.items()}
                params = self.marginals[m][1].predict(cur_x)
                local_pred[m].append(params)

        # merge batch-wise parameter dicts for each marginal and return
        results = []
        for m in range(n_marginals):
            parts = local_pred[m]
            keys = list(parts[0].keys())
            results.append({k: torch.cat([d[k] for d in parts]).detach().cpu().numpy() for k in keys})

        return results


def _var_indices(sel, adata: AnnData) -> np.ndarray:
    """Return integer indices of `sel` within `adata.var_names`.

    Expected use: `sel` is a list (or tuple) of variable names (strings).
    """
    # If sel is a single string, make it a list so we return consistent shape
    single_string = False
    if isinstance(sel, str):
        sel = [sel]
        single_string = True

    idx = np.asarray(adata.var_names.get_indexer(sel), dtype=int)
    if (idx < 0).any():
        missing = [s for s, i in zip(sel, idx) if i < 0]
        raise KeyError(f"Variables not found in adata.var_names: {missing}")
    return idx if not single_string else idx.reshape(-1)


from .copula import Copula
from .loader import obs_loader, adata_loader
from .marginal import Marginal
from .simulator import Simulator
from anndata import AnnData
from tqdm import tqdm
import torch
import numpy as np

class SCD3Simulator(Simulator):
    """Simulation wrapper"""

    def __init__(self, marginal: Marginal, copula: Copula):
        self.marginal = marginal
        self.copula = copula
        self.template = None
        self.parameters = None

    def fit(
        self,
        adata: AnnData,
        **kwargs):
        """Fit the simulator"""
        self.template = adata
        self.marginal.setup_data(adata, **kwargs)
        self.marginal.setup_optimizer(**kwargs)
        self.marginal.fit(**kwargs)

        # copula simulator
        self.copula.setup_data(adata, self.marginal.formula, **kwargs)
        self.copula.fit(self.marginal.uniformize, **kwargs)
        self.parameters = {
            "marginal": self.marginal.parameters,
            "copula": self.copula.parameters
        }

    def predict(self, obs=None, batch_size: int = 1000, **kwargs):
        """Predict from an obs dataframe"""
        # prepare an internal data loader for this obs
        if obs is None:
            obs = self.template.obs
        loader = obs_loader(
            obs,
            self.marginal.formula,
            batch_size=batch_size,
            **kwargs
        )

        # get predictions across batches
        local_parameters = []
        for _, x_dict in loader:
            l = self.marginal.predict(x_dict)
            local_parameters.append(l)

        # convert to a merged dictionary
        keys = list(local_parameters[0].keys())
        return {
            k: torch.cat([d[k] for d in local_parameters]).detach().cpu().numpy()
            for k in keys
        }

    def sample(self, obs=None, batch_size: int = 1000, **kwargs):
        """Generate samples."""
        if obs is None:
            obs = self.template.obs
        loader = obs_loader(
            obs,
            self.copula.formula | self.marginal.formula,
            batch_size=batch_size,
            **kwargs
        )

        # get samples across batches
        samples = []
        for _, x_dict in loader:
            u = self.copula.pseudo_obs(x_dict)
            u = torch.from_numpy(u)
            samples.append(self.marginal.invert(u, x_dict))
        samples = torch.cat(samples).detach().cpu().numpy()
        return AnnData(X = samples, obs=obs)

    def complexity(self, adata: AnnData = None, **kwargs):
        if adata is None:
            adata = self.template

        N, ll = 0, 0
        loader = adata_loader(adata, self.marginal.formula | self.copula.formula, **kwargs)
        for batch in tqdm(loader, desc="Computing log-likelihood..."):
            ll += self.copula.likelihood(self.marginal.uniformize, batch).sum()
            N += len(batch[0])

        return {
            "aic": -2 * ll + 2 * self.copula.num_params(),
            "bic": -2 * ll + np.log(N) * self.copula.num_params()
        }


from anndata import AnnData
from .. import format
from .. import data
from . import glm_factory as factory
import pandas as pd
import torch


def zero_inflated_poisson_regression_likelihood(params, X, y):
    # get appropriate parameter shape
    mean_n_features = X['mean'].shape[1]
    zero_inflation_n_features = X['zero_inflation'].shape[1]
    n_outcomes = y.shape[1]

    # define the likelihood parameters
    b_elem = mean_n_features * n_outcomes
    coef_mean = params[:b_elem].reshape(mean_n_features, n_outcomes)
    coef_zero_inflation = params[b_elem:].reshape(zero_inflation_n_features, n_outcomes)

    zero_inflation = torch.sigmoid(X['zero_inflation'] @ coef_zero_inflation)
    mu = torch.exp(X['mean'] @ coef_mean)
    poisson_loglikelihood = y * torch.log(mu + 1e-10) - mu - torch.lgamma(y + 1)

    # return the mixture, with an offset to prevent log(0)
    log_likelihood = torch.log(
        zero_inflation * (y == 0) + (1 - zero_inflation) * torch.exp(poisson_loglikelihood) + 1e-10
    )
    return -torch.sum(log_likelihood)


def zero_inflated_poisson_initializer(x, y, device):
    mean_n_features = x['mean'].shape[1]
    zero_inflation_n_features = x['zero_inflation'].shape[1]
    n_outcomes = y.shape[1]
    return torch.zeros(
        mean_n_features * n_outcomes + zero_inflation_n_features * n_outcomes, requires_grad=True, device=device
    )


def zero_inflated_poisson_postprocessor(params, x, y):
    mean_n_features = x['mean'].shape[1]
    zero_inflation_n_features = x['zero_inflation'].shape[1]
    n_outcomes = y.shape[1]
    b_elem = mean_n_features * n_outcomes
    coef_mean = format.to_np(params[:b_elem]).reshape(mean_n_features, n_outcomes)
    coef_zero_inflation = format.to_np(params[b_elem:]).reshape(zero_inflation_n_features, n_outcomes)
    return {"coef_mean": coef_mean, "coef_zero_inflation": coef_zero_inflation}


zero_inflated_poisson_regression_array = factory.multiple_formula_regression_factory(
    zero_inflated_poisson_regression_likelihood,
    zero_inflated_poisson_initializer,
    zero_inflated_poisson_postprocessor,
)


###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################


def format_zero_inflated_poisson_parameters(
    parameters: dict, var_names: list, mean_coef_index: list, 
    zero_inflation_coef_index: list
) -> dict:
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=mean_coef_index
    )
    parameters["coef_zero_inflation"] = pd.DataFrame(
        parameters["coef_zero_inflation"], columns=var_names, index=zero_inflation_coef_index
    )
    return parameters


def zero_inflated_poisson_regression(
    adata: AnnData, formula: str, chunk_size: int = int(1e4), batch_size=512, **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean', 'zero_inflation'])
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )
    parameters = zero_inflated_poisson_regression_array(loaders, **kwargs)
    return format_zero_inflated_poisson_parameters(
        parameters, list(adata.var_names), loaders["mean"].dataset.x_names, loaders["zero_inflation"].dataset.x_names
    )


import anndata
import torch
import torch.utils.data as td

class SparseMatrixLoader:
    def __init__(self, adata: anndata.AnnData, batch_size: int = None):
        ds = SparseMatrixDataset(adata, batch_size)
        self.loader = td.DataLoader(ds, batch_size=None)


class SparseMatrixDataset(td.IterableDataset):
    def __init__(self, anndata: anndata.AnnData, batch_size: int = None):
        self.n_rows = anndata.X.shape[0]
        if batch_size is None:
            batch_size = self.n_rows

        self.sparse_matrix = anndata.X
        self.batch_size = batch_size

    def __iter__(self):
        for i in range(0, self.n_rows, self.batch_size):
            batch_indices = range(i, min(i + self.batch_size, self.n_rows))
            batch_rows = self.sparse_matrix[batch_indices, :]

            # Convert to sparse CSR tensor
            batch_indices_rows, batch_indices_cols = batch_rows.nonzero()
            batch_values = batch_rows.data

            batch_sparse_tensor = torch.sparse_coo_tensor(
                torch.tensor([batch_indices_rows, batch_indices_cols]),
                torch.tensor(batch_values, dtype=torch.float32),
                (len(batch_indices), self.sparse_matrix.shape[1]),
            ).to_sparse_csr()

            yield batch_sparse_tensor

    def __len__(self):
        return (self.n_rows + self.batch_size - 1) // self.batch_size




from typing import Union
import warnings

def standardize_formula(formula: Union[str, dict], allowed_keys = None):
    # The first element of allowed_keys should be the name of default parameter
    if allowed_keys is None:
        raise ValueError("Internal error: allowed_keys must be specified")
    formula = {allowed_keys[0]: formula} if isinstance(formula, str) else formula

    formula_keys = set(formula.keys())
    allowed_keys = set(allowed_keys)

    if not formula_keys & allowed_keys:
        raise ValueError(f"formula must have at least one of the following keys: {allowed_keys}")
    
    if extra_keys := formula_keys - allowed_keys:
        warnings.warn(
            f"Invalid formulas in dictionary will not be used: {extra_keys}",
            UserWarning,
        )
    
    formula.update({k: '~ 1' for k in allowed_keys - formula_keys})
    return formula


from . import formula as fl
from anndata import AnnData
from formulaic import model_matrix
import numpy as np
import pandas as pd
import scipy
import torch
import torch.utils.data as td


def formula_group_loader(
    adata: AnnData,
    formula=None,
    grouping_variable=None,
    chunk_size=int(1e4),
    batch_size: int = None,
):
    device = fl.check_device()
    if grouping_variable is None:
        adata.obs["_copula_group"] = "shared_group"
        grouping_variable = "_copula_group"
        adata.obs["_copula_group"] = adata.obs["_copula_group"].astype("category")

    if adata.isbacked:
        ds = FormulaGroupViewDataset(
            adata, formula, grouping_variable, chunk_size, device
        )
        dataloader = td.DataLoader(ds, batch_size=batch_size, collate_fn=stack_collate())
        ds.x_names = fl.model_matrix_names(adata, formula, ds.categories)
    else:
        # convert sparse to dense matrix
        y = adata.X
        if isinstance(y, scipy.sparse._csc.csc_matrix):
            y = y.todense()

        # wrap the entire data into a dataset
        x = model_matrix(formula, pd.DataFrame(adata.obs))
        ds = td.StackDataset(
            x=td.TensorDataset(
                torch.tensor(np.array(x), dtype=torch.float32).to(device)
            ),
            y=td.TensorDataset(torch.tensor(y, dtype=torch.float32).to(device)),
            groups=ListDataset(adata.obs[grouping_variable]),
        )
        ds.groups = list(adata.obs[grouping_variable].dtype.categories)
        ds.x_names = list(x.columns)
        dataloader = td.DataLoader(ds, batch_size=batch_size, collate_fn=stack_collate(pop=True))

    return dataloader

def multiple_formula_group_loader(adata: AnnData, formulas: dict, grouping_variable=None,
                                  chunk_size=int(1e4), batch_size: int = None):
    dataloaders = {}
    for key in formulas.keys():
        dataloaders[key] = formula_group_loader(adata, formulas[key], grouping_variable, chunk_size, batch_size)
    return dataloaders 

class FormulaGroupViewDataset(td.Dataset):
    def __init__(
        self,
        view,
        formula=None,
        grouping_variable=None,
        chunk_size=int(1e4),
        device=None,
    ):
        super().__init__()
        self.device = device or fl.check_device()
        self.formula = formula
        self.categories = fl.column_levels(view.obs)
        self.grouping_variable = grouping_variable
        self.groups = list(self.categories[grouping_variable].dtype.categories)
        self.len = len(view)
        self.memberships = None
        self.view = view
        self.x = None
        self.y = None
        self.cur_range = range(0, min(self.len, chunk_size))

    def __len__(self):
        return self.len

    def __getitem__(self, ix):
        if self.x is None or ix not in self.cur_range:
            self.cur_range = range(ix, min(ix + len(self.cur_range), self.len))
            view_inmem = self.view[self.cur_range].to_memory()
            self.memberships = view_inmem.obs[self.grouping_variable]
            self.x = fl.safe_model_matrix(
                view_inmem.obs, self.formula, self.categories
            ).to(self.device)
            self.y = torch.from_numpy(view_inmem.X.toarray().astype(np.float32)).to(
                self.device
            )
        return {
            "x": self.x[ix - self.cur_range[0]],
            "y": self.y[ix - self.cur_range[0]],
            "groups": self.memberships[ix - self.cur_range[0]],
        }


class ListDataset(td.Dataset):
    """
    Simple DS to store groups
    """

    def __init__(self, list):
        self.list = list

    def __len__(self):
        return len(self.list)

    def __getitem__(self, idx):
        return self.list[idx]

def stack_collate(pop=False, groups=True):
    def f(batch):
        x = torch.stack([sample["x"][0] if pop else sample["x"] for sample in batch])
        y = torch.stack([sample["y"][0] if pop else sample["y"] for sample in batch])
        if groups:
            G = tuple([sample["groups"] for sample in batch])
            return [x, y, G]
        return [x, y]
    return f


import numpy as np
import pandas as pd
import anndata as ad
from typing import Union
from scipy.stats import norm


def glm_sample_factory(sample_array):
    def sampler(local_parameters: dict, obs: pd.DataFrame) -> ad.AnnData:
        samples = sample_array(local_parameters)
        result = ad.AnnData(X=samples, obs=obs)
        result.var_names = local_parameters["mean"].columns
        return result
    return sampler

def gaussian_copula_pseudo_obs(N, G, sigma, groups):
    u = np.zeros((N, G))

    # cycle across groups
    for group, ix in groups.items():
        if type(sigma) is not dict:
            sigma = {group: sigma}

        z = np.random.multivariate_normal(
            mean=np.zeros(G), cov=sigma[group], size=len(ix)
        )
        normal_distn = norm(0, np.diag(sigma[group] ** 0.5))
        u[ix] = normal_distn.cdf(z)
    return u


def gaussian_copula_sample_factory(copula_sample_array):
    def sampler(
        local_parameters: dict, covariance: Union[dict, np.array], groups: dict, obs: pd.DataFrame
    ) -> ad.AnnData:
        samples = copula_sample_array(local_parameters, covariance, groups)
        result = ad.AnnData(X=samples, obs=obs)
        result.var_names = local_parameters["mean"].columns
        return result
    return sampler




import numpy as np
from copy import deepcopy
from typing import Union


def substitute(
    params: dict, id: str, new_value: np.array, mask: Union[np.array, None] = None
) -> dict:
    if mask is None:
        mask = np.ones(params[id].shape)

    result = deepcopy(params)
    result[id][mask] = new_value
    return result



import numpy as np
import pandas as pd
from ..format import format_matrix
from typing import Union

def negbin_predict(parameters: dict, obs: pd.DataFrame, formula: Union[str, dict]):
    # Standardize formula to dictionary format
    if isinstance(formula, str):
        formula = {'mean': formula, 'dispersion': '~ 1'}
    
    x_mean = format_matrix(obs, formula["mean"]) 
    x_dispersion = format_matrix(obs, formula["dispersion"])
    
    r = np.exp(x_dispersion @ parameters["coef_dispersion"])
    mu = np.exp(x_mean @ parameters["coef_mean"])
    # r and mu are still dataframes with column names being the gene names
    return {"mean": mu, "dispersion": r}



from . import gaussian_copula_factory as gcf
from . import glm_factory as factory
from .. import format
from .. import data
from anndata import AnnData
from scipy.stats import norm
import numpy as np
import pandas as pd
import torch
from typing import Union

###############################################################################
## Regression functions that operate on numpy arrays
###############################################################################



# Gaussian regression likelihood: regression onto mean and sdev
def gaussian_regression_likelihood(params, X_dict, y):
    n_mean_features = X_dict["mean"].shape[1]
    n_sdev_features = X_dict["sdev"].shape[1]
    n_outcomes = y.shape[1]

    coef_mean = params[: n_mean_features * n_outcomes].reshape(n_mean_features, n_outcomes)
    coef_sdev = params[n_mean_features * n_outcomes :].reshape(n_sdev_features, n_outcomes)
    mu = X_dict["mean"] @ coef_mean
    sigma = torch.exp(X_dict["sdev"] @ coef_sdev)

    # Negative log likelihood for Gaussian
    log_likelihood = -0.5 * (torch.log(2 * torch.pi * sigma ** 2) + ((y - mu) ** 2) / (sigma ** 2))
    return torch.sum(log_likelihood)



def gaussian_initializer(x_dict, y, device):
    n_mean_features = x_dict["mean"].shape[1]
    n_outcomes = y.shape[1]
    n_sdev_features = x_dict["sdev"].shape[1]
    return torch.zeros(
        n_mean_features * n_outcomes + n_sdev_features * n_outcomes,
        requires_grad=True, device=device
    )



def gaussian_postprocessor(params, x_dict, y):
    n_mean_features = x_dict["mean"].shape[1]
    n_outcomes = y.shape[1]
    n_sdev_features = x_dict["sdev"].shape[1]
    coef_mean = format.to_np(params[:n_mean_features * n_outcomes]).reshape(n_mean_features, n_outcomes)
    coef_sdev = format.to_np(params[n_mean_features * n_outcomes:]).reshape(n_sdev_features, n_outcomes)
    return {"coef_mean": coef_mean, "coef_sdev": coef_sdev}



gaussian_regression_array = factory.multiple_formula_regression_factory(
    gaussian_regression_likelihood, gaussian_initializer, gaussian_postprocessor
)


###############################################################################
## Regression functions that operate on AnnData objects
###############################################################################


def format_gaussian_parameters(
    parameters: dict, var_names: list, mean_coef_index: list, sdev_coef_index: list
) -> dict:
    parameters["coef_mean"] = pd.DataFrame(
        parameters["coef_mean"], columns=var_names, index=mean_coef_index
    )
    parameters["coef_sdev"] = pd.DataFrame(
        parameters["coef_sdev"], columns=var_names, index=sdev_coef_index
    )
    return parameters


def format_gaussian_parameters_with_loaders(
    parameters: dict, var_names: list, dls: dict
) -> dict:
    mean_coef_index = dls["mean"].dataset.x_names
    sdev_coef_index = dls["sdev"].dataset.x_names
    return format_gaussian_parameters(
        parameters, var_names, mean_coef_index, sdev_coef_index
    )


def gaussian_regression(
    adata: AnnData, formula: Union[str, dict], chunk_size: int = int(1e4), 
    batch_size=512, **kwargs
) -> dict:
    formula = data.standardize_formula(formula, allowed_keys=['mean', 'sdev'])
    loaders = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )
    parameters = gaussian_regression_array(loaders, **kwargs)
    return format_gaussian_parameters(
        parameters, list(adata.var_names), loaders["mean"].dataset.x_names, 
        loaders["sdev"].dataset.x_names
    )


###############################################################################
## Copula versions for gaussian regression
###############################################################################

def gaussian_uniformizer(parameters, X_dict, y, epsilon=1e-3):
    mu = X_dict["mean"] @ parameters["coef_mean"]
    sigma = np.exp(X_dict["sdev"] @ parameters["coef_sdev"])
    u = norm.cdf(y, loc=mu, scale=sigma)
    u = np.clip(u, epsilon, 1 - epsilon)
    return u

gaussian_copula_array = gcf.gaussian_copula_array_factory(
    gaussian_regression_array, gaussian_uniformizer
)

gaussian_copula = gcf.gaussian_copula_factory(
    gaussian_copula_array, format_gaussian_parameters_with_loaders,
    param_name=['mean', 'sdev']
)



from scdesigner.margins.negative_binomial import NegativeBinomial
from scdesigner.simulator import simulator
from scipy.sparse import csr_matrix
import anndata as ad
import numpy as np
import pandas as pd


def generate_adata(N=100, G=2000):
    """
    Helper function used in tests
    """
    counts = csr_matrix(np.random.poisson(1, size=(100, 2000)), dtype=np.float32)
    adata = ad.AnnData(counts)
    ct = np.random.choice(["B", "T", "Monocyte"], size=(adata.n_obs,))
    adata.obs["cell_type"] = pd.Categorical(ct)
    adata.var_names = [f"Gene_{i:d}" for i in range(adata.n_vars)]
    return adata


def test_nb_simulator():
    adata = generate_adata()
    sim = simulator(adata, NegativeBinomial("~ cell_type"))
    genes, params = sim.parameters()
    assert all(genes == adata.var_names)
    assert params["B"].shape == (3, adata.n_vars)
    assert params["A"].shape == (1, adata.n_vars)


def test_nb_predict():
    adata = generate_adata()
    sim = simulator(adata, NegativeBinomial("~ cell_type"))
    y_hat = sim.predict(adata.obs)
    assert np.all(np.sort(list(y_hat.keys())) == ["alpha", "mu"])
    assert y_hat["alpha"].shape == (adata.n_obs, adata.n_vars)
    assert y_hat["mu"].shape == (adata.n_obs, adata.n_vars)
    assert np.all(y_hat["mu"] > 0)
    assert np.all(y_hat["alpha"] > 0)



from scipy.stats import nbinom
from . import glm_factory as glm
from typing import Union
import numpy as np


def negbin_regression_sample_array(local_parameters: dict) -> np.array:
    r, mu = local_parameters["dispersion"], local_parameters["mean"] # dataframes of shape (n, g)
    return nbinom(n=r, p=r / (r + mu)).rvs()


def negbin_copula_sample_array(
    local_parameters: dict, covariance: Union[dict, np.array], groups: dict
) -> np.array:
    # initialize uniformized gaussian samples
    N, G = local_parameters["mean"].shape
    u = glm.gaussian_copula_pseudo_obs(N, G, covariance, groups)

    # transform the correlated uniforms to NB space
    r, mu = local_parameters["dispersion"], local_parameters["mean"]
    return nbinom(n=r, p=r / (r + mu)).ppf(u)


negbin_sample = glm.glm_sample_factory(negbin_regression_sample_array)
negbin_copula_sample = glm.gaussian_copula_sample_factory(negbin_copula_sample_array)



from .formula import standardize_formula
from .marginal import GLMPredictor, Marginal
from .loader import _to_numpy
from typing import Union, Dict, Optional
import torch
import numpy as np
from scipy.stats import nbinom

class NegBin(Marginal):
    """Negative-binomial marginal estimator"""
    def __init__(self, formula: Union[Dict, str]):
        formula = standardize_formula(formula, allowed_keys=['mean', 'dispersion'])
        super().__init__(formula)

    def setup_optimizer(
            self,
            optimizer_class: Optional[callable] = torch.optim.Adam,
            **optimizer_kwargs,
    ):
        if self.loader is None:
            raise RuntimeError("self.loader is not set (call setup_data first)")

        nll = lambda batch: -self.likelihood(batch).sum()
        self.predict = GLMPredictor(
            n_outcomes=self.n_outcomes,
            feature_dims=self.feature_dims,
            loss_fn=nll,
            optimizer_class=optimizer_class,
            optimizer_kwargs=optimizer_kwargs
        )

    def likelihood(self, batch):
        """Compute the negative log-likelihood"""
        y, x = batch
        params = self.predict(x)
        mu = params.get('mean')
        r = params.get('dispersion')
        return (
            torch.lgamma(y + r)
            - torch.lgamma(r)
            - torch.lgamma(y + 1.0)
            + r * torch.log(r)
            + y * torch.log(mu)
            - (r + y) * torch.log(r + mu)
        )

    def invert(self, u: torch.Tensor, x: Dict[str, torch.Tensor]):
        """Invert pseudoobservations."""
        mu, r, u = self._local_params(x, u)
        p = r / (r + mu)
        y = nbinom(n=r, p=p).ppf(u)
        return torch.from_numpy(y).float()

    def uniformize(self, y: torch.Tensor, x: Dict[str, torch.Tensor], epsilon=1e-6):
        """Return uniformized pseudo-observations for counts y given covariates x."""
        # cdf values using scipy's parameterization
        mu, r, y = self._local_params(x, y)
        p = r / (r + mu)
        u1 = nbinom(n=r, p=p).cdf(y)
        u2 = np.where(y > 0, nbinom(n=r, p=p).cdf(y - 1), 0.0)

        # randomize within discrete mass to get uniform(0,1)
        v = np.random.uniform(size=y.shape)
        u = np.clip(v * u1 + (1.0 - v) * u2, epsilon, 1.0 - epsilon)
        return torch.from_numpy(u).float()

    def _local_params(self, x, y=None):
        params = self.predict(x)
        mu = params.get('mean')
        r = params.get('dispersion')
        if y is None:
            return _to_numpy(mu, r)
        return _to_numpy(mu, r, y)



from .. import data
from anndata import AnnData
from formulaic import model_matrix
from scipy.stats import norm, multivariate_normal
from typing import Union
import numpy as np
import pandas as pd
import torch, scipy


def marginal_aic_bic(likelihood, params: dict, adata: AnnData, 
                     formula: Union[str, dict], allowed_keys: set,
                     param_order: list = None, transform: list = None, 
                     chunk_size: int = int(1e4), batch_size=512):
    device = data.formula.check_device()
    nsample = len(adata)
    params = likelihood_unwrapper(params, param_order, transform).to(device)
    nparam = len(params)

    # create batches for likelihood calculation
    formula = data.standardize_formula(formula, allowed_keys)
    loader = data.multiple_formula_loader(
        adata, formula, chunk_size=chunk_size, batch_size=batch_size
    )
    keys = list(loader.keys())
    loaders = list(loader.values())
    num_keys = len(keys)
    
    ll = 0
    with torch.no_grad():
        for batches in zip(*loaders):
            x = {keys[i]: batches[i][0].to(device) for i in range(num_keys)}
            y = batches[0][1].to(device)
            ll += -likelihood(params, x, y)
    aic = 2 * nparam - 2 * ll
    bic = np.log(nsample) * nparam - 2 * ll
    return aic.cpu().item(), bic.cpu().item()


def gaussian_copula_aic_bic(uniformizer, params: dict, adata: AnnData, 
                            formula: Union[str, dict], allowed_keys: set, copula_groups=None):
    params = uniformizer_unwrapper(params)
    covariance = params['covariance']
    y = adata.X
    if isinstance(y, scipy.sparse._csc.csc_matrix):
        y = y.todense()
    formula = data.standardize_formula(formula, allowed_keys)
    X = {key: model_matrix(formula[key], pd.DataFrame(adata.obs)) for key in formula}
    if copula_groups is not None:
        memberships = adata.obs[copula_groups]
    else:
        copula_groups = "shared_group"
        memberships = np.array(["shared_group"] * y.shape[0])
        
    u = uniformizer(params, X, y)
    groups = covariance.keys()
    nparam = {
        g: (np.sum(covariance[g] != 0) - covariance[g].shape[0]) / 2 for g in groups
    }
    aic = 0 # in the future may add group-wise AIC/BIC
    bic = 0
    for g in groups:
        ix = np.where(memberships == g)[0]
        z = norm().ppf(u[ix])
        copula_ll = multivariate_normal.logpdf(
            z, np.zeros(covariance[g].shape[0]), covariance[g]
        )
        marginal_ll = norm.logpdf(z)
        aic += -2 * (np.sum(copula_ll) - np.sum(marginal_ll)) + 2 * nparam[g]
        bic += (
            -2 * (np.sum(copula_ll) - np.sum(marginal_ll))
            + np.log(z.shape[0]) * nparam[g]
        )
    return aic, bic


def compose_marginal_diagnose(likelihood, allowed_keys: set, param_order: list = None, transform: list = None):
    def diagnose(params: dict, adata: AnnData, formula: Union[str, dict],
                chunk_size: int = int(1e4), batch_size=512):
        return marginal_aic_bic(likelihood, params, adata, formula, allowed_keys,
                                param_order, transform, chunk_size, batch_size)
    return diagnose


def compose_gcopula_diagnose(likelihood, uniformizer, allowed_keys: set, 
                             param_order: list = None, transform: list = None):
    def diagnose(params: dict, adata: AnnData, formula: str, copula_groups=None,
                 chunk_size: int = int(1e4), batch_size=512):
        marginal_aic, marginal_bic = marginal_aic_bic(likelihood, params, adata, formula, allowed_keys,
                                                    param_order, transform, chunk_size, batch_size)
        copula_aic, copula_bic = gaussian_copula_aic_bic(uniformizer, params, adata, formula, allowed_keys, copula_groups)
        return marginal_aic, marginal_bic, copula_aic, copula_bic
    return diagnose


###############################################################################
## Helper for converting params to match likelihood/uniformizer input format
###############################################################################

def likelihood_unwrapper(params: dict, param_order: list = None, transform: list = None):
    l = []
    keys_to_process = [k for k in params if k != "covariance"] if param_order is None else param_order

    for idx, k in enumerate(keys_to_process):
        feature = params[k]
        v = torch.Tensor(feature.values).reshape(1, feature.shape[0] * feature.shape[1])[0]
        if transform is not None:
            v = transform[idx](v)
        l.append(v)

    return torch.cat(l, dim=0)

def uniformizer_unwrapper(params):
    params = params = {key: params[key].values if key!='covariance' else params[key] for key in params}
    if not isinstance(params['covariance'], dict):
        params['covariance'] = {'shared_group': params['covariance'].values}
    else:
        params['covariance'] = {key: params['covariance'][key].values for key in params['covariance']}
    return params



import numpy as np
from typing import Union
from copy import deepcopy


def amplify(
    params: dict, id: str, mask: Union[np.array, None] = None, factor: float = 1
) -> dict:
    if mask is None:
        mask = np.ones(params[id].shape)

    result = deepcopy(params)
    result[id].values[mask] = factor * result[id].values[mask]
    return result



from scdesigner.margins.negative_binomial import NegativeBinomial
import numpy as np
import pandas as pd
import torch
import torch.optim
import torch.random


def test_nb_mean():
    # define ground truth mean
    N, G = 1000, 20
    alpha = np.random.uniform(2.5, 5, G)
    mu = np.ones((N, 1)) @ np.exp(np.random.normal(size=(1, G)))

    # generate samples
    Y = np.random.negative_binomial(1 / alpha, 1 / (1 + alpha * mu))
    Y = torch.from_numpy(Y)

    # estimate means
    nb_model = NegativeBinomial("~ 1")
    nb_model.fit(Y)
    mu_hat = nb_model.parameters["B"].values
    assert np.mean(np.abs(mu_hat - np.log(mu[0]))) < 0.1


def test_mu_regression():
    # define ground truth coefficients
    N, G, D = 1000, 20, 4
    alpha = np.random.uniform(2.5, 5, G)
    X = np.random.normal(size=(N, D))
    B = np.random.normal(size=(D, G))
    mu = np.exp(X @ B)

    # generate samples
    Y = np.random.negative_binomial(1 / alpha, 1 / (1 + alpha * mu))
    Y = torch.from_numpy(Y)

    # estimate means
    X = pd.DataFrame(X, columns=[f"dim{j}" for j in range(D)])
    nb_model = NegativeBinomial("~ . - 1")
    nb_model.fit(Y, X, max_iter=50)
    Bhat = nb_model.parameters["B"].values
    assert np.mean(np.abs(Bhat - B)) < 0.1


def test_nb_dispersion():
    # define ground truth dispersion
    N, G = 1000, 20
    mu = np.random.uniform(2.5, 5, G)
    alpha = np.ones((N, 1)) @ np.exp(np.random.normal(size=(1, G)))

    # generate samples
    Y = np.random.negative_binomial(1 / alpha, 1 / (1 + alpha * mu))
    Y = torch.from_numpy(Y)

    # estimate dispersion parameters
    nb_model = NegativeBinomial({"alpha": "~ 1"})
    nb_model.fit(Y)
    Ahat = nb_model.parameters["A"].values
    assert np.mean(np.abs(Ahat - np.log(alpha[0]))) < 0.1


def test_alpha_regression():
    # define ground truth coefficients
    N, G, D = 1000, 20, 4
    mu = np.random.uniform(2, 5, G)
    X = np.random.normal(size=(N, D))
    X = pd.DataFrame(X, columns=[f"dim{j}" for j in range(D)])
    A = np.random.normal(size=(D, G))
    alpha = np.exp(X @ A)

    # generate samples
    Y = np.random.negative_binomial(1 / alpha, 1 / (1 + alpha * mu))
    Y = torch.from_numpy(Y)

    # estimate means
    nb_model = NegativeBinomial({"alpha": "~ . - 1"})
    nb_model.fit(Y, pd.DataFrame(X))
    Ahat = nb_model.parameters["A"].values
    assert np.mean(np.abs(Ahat - A)) < 0.1



from . import glm_factory as glm
from scipy.stats import bernoulli
from typing import Union
import numpy as np


def bernoulli_regression_sample_array(local_parameters: dict) -> np.array:
    theta = local_parameters["mean"]
    return bernoulli(theta).rvs()


def bernoulli_copula_sample_array(
    local_parameters: dict, covariance: Union[dict, np.array], groups: dict
) -> np.array:
    # initialize uniformized gaussian samples
    N, G = local_parameters["mean"].shape
    u = glm.gaussian_copula_pseudo_obs(N, G, covariance, groups)

    theta = local_parameters["mean"]
    return bernoulli(theta).ppf(u)


bernoulli_sample = glm.glm_sample_factory(bernoulli_regression_sample_array)

bernoulli_copula_sample = glm.gaussian_copula_sample_factory(
    bernoulli_copula_sample_array
)



from scipy.stats import nbinom, bernoulli
from . import glm_factory as glm
from typing import Union
import numpy as np


def zero_inflated_negbin_sample_array(local_parameters: dict) -> np.array:
    mu, r, pi = (
        local_parameters["mean"],
        local_parameters["dispersion"],
        local_parameters["zero_inflation"],
    )
    return nbinom(n=r, p=r / (r + mu)).rvs() * bernoulli(1 - pi).rvs()


def zero_inflated_negbin_copula_sample_array(
    local_parameters: dict, covariance: Union[dict, np.array], groups: dict
) -> np.array:
    # initialize uniformized gaussian samples
    N, G = local_parameters["mean"].shape
    u = glm.gaussian_copula_pseudo_obs(N, G, covariance, groups)

    # get zero inflated NB parameters
    mu, r, pi = (
        local_parameters["mean"],
        local_parameters["dispersion"],
        local_parameters["zero_inflation"],
    )

    # zero inflate after first simulating from NB
    positive_part = nbinom(n=r, p=r / (r + mu)).ppf(u)
    zero_inflation = bernoulli(1 - pi).ppf(u)
    return zero_inflation * positive_part


zero_inflated_negbin_sample = glm.glm_sample_factory(zero_inflated_negbin_sample_array)

zero_inflated_negbin_copula_sample = glm.gaussian_copula_sample_factory(
    zero_inflated_negbin_copula_sample_array
)



import scanpy as sc
import numpy as np
import pandas as pd
import altair as alt
import matplotlib.pyplot as plt


def adata_df(adata):
    return (
        pd.DataFrame(adata.X, columns=adata.var_names)
        .melt(id_vars=[], value_vars=adata.var_names)
        .reset_index(drop=True)
    )


def merge_samples(adata, sim):
    source = adata_df(adata)
    simulated = adata_df(sim)
    return pd.concat(
        {"real": source, "simulated": simulated}, names=["source"]
    ).reset_index(level="source")


def plot_umap(
    adata,
    color=None,
    shape=None,
    facet=None,
    opacity=0.6,
    n_comps=20,
    n_neighbors=15,
    transform=lambda x: np.log1p(x),
    **kwargs
):
    mapping = {"x": "UMAP1", "y": "UMAP2", "color": color, "shape": shape}
    mapping = {k: v for k, v in mapping.items() if v is not None}

    adata_ = adata.copy()
    adata_.X = check_sparse(adata_.X)
    Z = transform(adata_.X)
    if Z.shape[1] == adata_.X.shape[1]:
        adata_.X = transform(adata_.X)
    else:
        adata_ = adata_[:, : Z.shape[1]]
        adata_.X = Z
        adata_.var_names = [f"transform_{k}" for k in range(Z.shape[1])]

    # umap on the top PCA dimensions
    sc.pp.pca(adata_, n_comps=n_comps)
    sc.pp.neighbors(adata_, n_neighbors=n_neighbors, n_pcs=n_comps)
    sc.tl.umap(adata_, **kwargs)

    # get umap embeddings
    umap_df = pd.DataFrame(adata_.obsm["X_umap"], columns=["UMAP1", "UMAP2"])
    umap_df = pd.concat([umap_df, adata_.obs.reset_index(drop=True)], axis=1)

    # encode and visualize
    chart = alt.Chart(umap_df).mark_point(opacity=opacity).encode(**mapping)
    if facet is not None:
        chart = chart.facet(column=alt.Facet(facet))
    return chart


def plot_pca(
    adata,
    color=None,
    shape=None,
    facet=None,
    opacity=0.6,
    plot_dims=[0, 1],
    transform=lambda x: np.log1p(x),
    **kwargs
):
    mapping = {"x": "PCA1", "y": "PCA2", "color": color, "shape": shape}
    mapping = {k: v for k, v in mapping.items() if v is not None}

    adata_ = adata.copy()
    adata_.X = check_sparse(adata_.X)
    adata_.X = transform(adata_.X)

    # get PCA scores
    sc.pp.pca(adata_, **kwargs)
    pca_df = pd.DataFrame(adata_.obsm["X_pca"][:, plot_dims], columns=["PCA1", "PCA2"])
    pca_df = pd.concat([pca_df, adata_.obs.reset_index(drop=True)], axis=1)

    # plot
    chart = alt.Chart(pca_df).mark_point(opacity=opacity).encode(**mapping)
    if facet is not None:
        chart = chart.facet(column=alt.Facet(facet))
    return chart


def compare_summary(real, simulated, summary_fun):
    df = pd.DataFrame({"real": summary_fun(real), "simulated": summary_fun(simulated)})

    identity = pd.DataFrame(
        {
            "real": [df["real"].min(), df["real"].max()],
            "simulated": [df["real"].min(), df["real"].max()],
        }
    )
    return alt.Chart(identity).mark_line(color="#dedede").encode(
        x="real", y="simulated"
    ) + alt.Chart(df).mark_circle().encode(x="real", y="simulated")


def check_sparse(X):
    if not isinstance(X, np.ndarray):
        X = X.todense()
    return X


def compare_means(real, simulated, transform=lambda x: x):
    real_, simulated_ = prepare_dense(real, simulated)
    summary = lambda a: np.asarray(transform(a.X).mean(axis=0)).flatten()
    return compare_summary(real_, simulated_, summary)


def prepare_dense(real, simulated):
    real_ = real.copy()
    simulated_ = simulated.copy()
    real_.X = check_sparse(real_.X)
    simulated_.X = check_sparse(simulated_.X)
    return real_, simulated_


def compare_variances(real, simulated, transform=lambda x: x):
    real_, simulated_ = prepare_dense(real, simulated)
    summary = lambda a: np.asarray(np.var(transform(a.X), axis=0)).flatten()
    return compare_summary(real_, simulated_, summary)


def compare_standard_deviation(real, simulated, transform=lambda x: x):
    real_, simulated_ = prepare_dense(real, simulated)
    summary = lambda a: np.asarray(np.std(transform(a.X), axis=0)).flatten()
    return compare_summary(real_, simulated_, summary)


def concat_real_sim(real, simulated):
    real_, simulated_ = prepare_dense(real, simulated)
    real_.obs["source"] = "real"
    simulated_.obs["source"] = "simulated"
    return real_.concatenate(simulated_, join="outer", batch_key=None)


def compare_umap(real, simulated, transform=lambda x: x, **kwargs):
    adata = concat_real_sim(real, simulated)
    return plot_umap(adata, facet="source", transform=transform, **kwargs)


def compare_pca(real, simulated, transform=lambda x: x, **kwargs):
    adata = concat_real_sim(real, simulated)
    return plot_pca(adata, facet="source", transform=transform, **kwargs)


def plot_hist(sim_data, real_data, idx):
    sim = sim_data[:, idx]
    real = real_data[:, idx]
    b = np.linspace(min(min(sim), min(real)), max(max(sim), max(real)), 50)

    plt.hist([real, sim], b, label=["Real", "Simulated"], histtype="bar")
    plt.xlabel("x")
    plt.ylabel("Density")
    plt.legend()
    plt.show()


def compare_ecdf(adata, sim, var_names=None, max_plot=10, n_cols=5, **kwargs):
    if var_names is None:
        var_names = adata.var_names[:max_plot]

    combined = merge_samples(adata[:, var_names], sim.sample()[:, var_names])
    alt.data_transformers.enable("vegafusion")

    plot = (
        alt.Chart(combined)
        .transform_window(
            ecdf="cume_dist()", sort=[{"field": "value"}], groupby=["variable"]
        )
        .mark_line(
            interpolate="step-after",
        )
        .encode(
            x="value:Q",
            y="ecdf:Q",
            color="source:N",
            facet=alt.Facet(
                "variable", sort=alt.EncodingSortField("value"), columns=n_cols
            ),
        )
        .properties(**kwargs)
    )
    plot.show()
    return plot, combined


def compare_boxplot(adata, sim, var_names=None, max_plot=20, **kwargs):
    if var_names is None:
        var_names = adata.var_names[:max_plot]

    combined = merge_samples(adata[:, var_names], sim.sample()[:, var_names])
    alt.data_transformers.enable("vegafusion")

    plot = (
        alt.Chart(combined)
        .mark_boxplot(extent="min-max")
        .encode(
            x=alt.X("value:Q").scale(zero=False),
            y=alt.Y(
                "variable:N",
                sort=alt.EncodingSortField("mid_box_value", order="descending"),
            ),
            facet="source:N",
        )
        .properties(**kwargs)
    )
    plot.show()
    return plot, combined


def compare_histogram(adata, sim, var_names=None, max_plot=20, n_cols=5, **kwargs):
    if var_names is None:
        var_names = adata.var_names[:max_plot]

    combined = merge_samples(adata[:, var_names], sim.sample()[:, var_names])
    alt.data_transformers.enable("vegafusion")

    plot = (
        alt.Chart(combined)
        .mark_bar(opacity=0.7)
        .encode(
            x=alt.X("value:Q").bin(maxbins=20),
            y=alt.Y("count()").stack(None),
            color="source:N",
            facet=alt.Facet(
                "variable", sort=alt.EncodingSortField("bin_maxbins_20_value")
            ),
        )
        .properties(**kwargs)
    )
    plot.show()
    return plot, combined



import numpy as np
import re


def str_match(string: str, string_list: list) -> bool:
    for l in string_list:
        if l in string:
            return True
    return False


def data_frame_mask(df, row_pattern=".", col_pattern=".") -> np.array:
    mask = np.zeros(df.shape, dtype=bool)
    if isinstance(col_pattern, str):
        col_pattern = [col_pattern]
    if isinstance(row_pattern, str):
        row_pattern = [row_pattern]

    # check for columns that match at least one pattern
    col_matches = np.zeros(df.shape[1], dtype=bool)
    for col in df.columns:
        if any(re.search(pat, col) for pat in col_pattern):
            col_matches[df.columns.get_loc(col)] = True

    # same with rows
    row_matches = np.zeros(df.shape[0], dtype=bool)
    for idx in df.index:
        if any(re.search(pat, str(idx)) for pat in row_pattern):
            row_matches[df.index.get_loc(idx)] = True

    # set mask to the intersection of row and column matches
    mask = np.outer(row_matches, col_matches)
    return mask



from typing import Dict, Callable, Tuple
import torch
from anndata import AnnData
from .loader import adata_loader

class Copula:
    def __init__(self, formula: str, **kwargs):
        self.formula = formula
        self.loader = None
        self.n_outcomes = None
        self.parameters = None

    def setup_data(self, adata: AnnData, marginal_formula: Dict[str, str], batch_size: int = 1024, **kwargs):
        self.adata = adata
        self.formula = self.formula | marginal_formula
        self.loader = adata_loader(adata, self.formula, batch_size=batch_size, **kwargs)
        X_batch, _ = next(iter(self.loader))
        self.n_outcomes = X_batch.shape[1]

    def fit(self, uniformizer: Callable, **kwargs):
        raise NotImplementedError

    def pseudo_obs(self, x_dict: Dict):
        raise NotImplementedError

    def likelihood(self, uniformizer: Callable, batch: Tuple[torch.Tensor, Dict[str, torch.Tensor]]):
        raise NotImplementedError

    def num_params(self, **kwargs):
        raise NotImplementedError

    def format_parameters(self):
        raise NotImplementedError